# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Functions to import the reconstruction part of colmap (only text format is supported).
"""

import logging
import numpy as np
import os.path as path
import re
from typing import Dict, Tuple, Optional
# kapture
import kapture
import kapture.io.features
# local
from .cameras import get_camera_kapture_id_from_colmap_id

logger = logging.getLogger('colmap')

colmap_reconstruction_split_pattern = r'[^,\s]+'


def import_from_colmap_cameras_txt(colmap_cameras_filepath: str) -> kapture.Sensors:
    """
    Imports Sensors from colmap cameras.txt

    :param colmap_cameras_filepath: input path to colmap cameras.txt file
    :return: kapture sensors
    """
    sensors = kapture.Sensors()
    # cameras[cam_id] = camera
    with open(colmap_cameras_filepath, 'r') as colmap_cameras_filepath:
        lines = colmap_cameras_filepath.readlines()
        # eliminate comments
        lines = (line for line in lines if not line.startswith('#'))
        # split by space and or comma
        lines = (re.findall(colmap_reconstruction_split_pattern, line.rstrip())
                 for line in lines)  # split fields
        for fields in lines:
            camera_id = get_camera_kapture_id_from_colmap_id(int(fields[0]))
            camera_type = str(fields[1])
            image_size = [str(s) for s in fields[2:4]]
            projection_params = [str(f) for f in fields[4:]]
            camera = kapture.Camera(camera_type, image_size + projection_params)
            sensors[camera_id] = camera
    return sensors


def import_from_colmap_images_txt(colmap_images_filepath: str,
                                  kapture_dirpath: Optional[str] = None,
                                  keypoints_type: str = 'SIFT'
                                  ) -> Tuple[kapture.RecordsCamera, kapture.Trajectories, Optional[kapture.Keypoints]]:
    """
    Imports RecordsCamera, Trajectories and Keypoints from colmap images.txt

    :param colmap_images_filepath: path to colmap images.txt file
    :param kapture_dirpath: path to kapture root path.
                            If not given (None), keypoints are not created.
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    (by default, in colmap, it is SIFT, but can be imported)
    :return: kapture images, trajectories and keypoints
    """

    # colmap images file format is :
    # Image list with two lines of data per image:
    #   IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME
    #   POINTS2D[] as (X, Y, POINT3D_ID)

    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories()
    keypoints = None
    image_names = []  # first pass
    # first pass: IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME
    # to images and trajectories
    with open(colmap_images_filepath, 'r') as colmap_images_file:
        lines = colmap_images_file.readlines()
        lines = (line for line in lines if not line.startswith('#'))  # eliminate comments
        lines = (line for i, line in enumerate(lines) if (i % 2) == 0)  # eliminate even lines
        # split by space and or comma
        lines = (re.findall(colmap_reconstruction_split_pattern, line.rstrip())
                 for line in lines)  # split fields
        # but make sure not to split spaces in file names
        lines = (line[0:9] + [' '.join(line[9:])] for line in lines)
        for fields in lines:
            # IMAGE_ID, QW, QX, QY, QZ, TX, TY, TZ, CAMERA_ID, NAME
            timestamp = int(fields[0])  # use image ID as timestamp
            q = [float(v) for v in fields[1:5]]
            t = [float(v) for v in fields[5:8]]
            pose = kapture.PoseTransform(q, t)
            camera_id = get_camera_kapture_id_from_colmap_id(int(fields[8]))
            image_name = fields[9]
            images[timestamp, camera_id] = image_name
            trajectories[timestamp, camera_id] = pose
            image_names.append(image_name)

    # second pass: keypoints, observations and points 3d
    if kapture_dirpath is not None:
        # second pass: POINTS2D[] as (X, Y, POINT3D_ID)
        image_names_with_keypoints = set()
        # observations = kapture.Observations()
        with open(colmap_images_filepath, 'r') as colmap_images_file:
            lines = colmap_images_file.readlines()
            lines = (line for line in lines if not line.startswith('#'))  # eliminate comments
            lines = (line for i, line in enumerate(lines) if (i % 2) == 1)  # eliminate odd lines
            # split by space and or comma
            lines = (re.findall(colmap_reconstruction_split_pattern, line.rstrip())
                     for line in lines)  # split fields
            for image_name, fields in zip(image_names, lines):
                image_keypoints_colmap = np.array(fields).reshape((-1, 3))[:, 0:2].astype(np.float32)
                # register as keypoints if there is at least one
                if image_keypoints_colmap.shape[0] > 0:
                    keypoints_filepath = kapture.io.features.get_keypoints_fullpath(keypoints_type,
                                                                                    kapture_dirpath,
                                                                                    image_name)
                    kapture.io.features.image_keypoints_to_file(keypoints_filepath, image_keypoints_colmap)
                    image_names_with_keypoints.add(image_name)
                    # TODO: observations

        if image_names_with_keypoints:
            keypoints = kapture.Keypoints(keypoints_type, np.float32, 2, image_names_with_keypoints)

    return images, trajectories, keypoints


def import_from_colmap_points3d_txt(colmap_points3d_filepath: str,
                                    image_names: Dict[int, str] = None,
                                    keypoints_type: str = 'SIFT',
                                    skip_observations: bool = False
                                    ) -> Tuple[kapture.Points3d, Optional[kapture.Observations]]:
    """
    Imports the colmap file named "points3d.txt" containing both points 3D and observations.

    :param colmap_points3d_filepath: path to the colmap file named "points3d.txt"
    :param image_names: dict of image names matching the colmap image id to the kapture image name
                        colmap_image_idx -> kapture_image_filename
    :param keypoints_type: type of keypoints, name of the keypoints subfolder
    (by default, in colmap, it is SIFT, but can be imported)
    :param skip_observations: skip import of observations id true.
    :return: kapture 3D points and observations
    """
    assert path.basename(colmap_points3d_filepath) == 'points3D.txt'
    points3d = []
    observations = kapture.Observations()

    # colmap points3D.txt contains both points 3D and observations
    with open(colmap_points3d_filepath, 'r') as file:
        # in the reconstruction, spaces and commas can be used as separators
        lines = file.readlines()
        # eliminate comments
        lines = (line for line in lines if not line.startswith('#'))
        # split by space and or comma
        lines = ([float(value) for value in re.findall(colmap_reconstruction_split_pattern, values)]
                 for values in lines)  # split into an array of floats
        for index, values in enumerate(lines):
            points3d.append(values[1:4] + values[4:7])
            if not skip_observations and len(values) > 8 and len(image_names) > 0:
                for image_idx, point_2d_idx in zip(values[8::2], values[9::2]):
                    filename = image_names.get(int(image_idx), 'unknown')
                    observations.add(index, keypoints_type, filename, int(point_2d_idx))

    points3d = kapture.Points3d(np.array(points3d)) if points3d else kapture.Points3d()
    observations = None if len(observations) == 0 else observations
    return points3d, observations



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Convert kapture data to Long-term Visual Localization challenge format

This line should store the
result as `name.jpg qw qx qy qz tx ty tz`.
Here,  `name` corresponds to the filename of the image, without any directory names.
`qw qx qy qz` represents the **rotation** from world to camera coordinates as a
**unit quaternion**. `tx ty tz` is the camera **translation** (**not the camera position**).

"""
import logging
import os
import os.path as path
import pathlib
import argparse
from tqdm import tqdm

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.features
import kapture.io.csv

logger = logging.getLogger('LTVL2020')


def export_ltvl(kapture_dirpath: str,
                ltvl_filepath: str,
                prepend_camera_name: bool = False,
                keep_full_file_name: bool = False) -> None:
    """
    Export kapture data to a Long-term Visual Localization challenge format file.

    :param kapture_dirpath: kapture data top directory
    :param ltvl_filepath: LTVL file path to write
    :param prepend_camera_name: if True, it will prepend the camera name to the image file names.
    """
    # only load (1) image records + (2) trajectories (that all that matters).
    # 1: load records
    records_camera_filepath = kapture.io.csv.get_csv_fullpath(kapture.RecordsCamera, kapture_dirpath)
    logger.debug(f'loading {records_camera_filepath}')
    records_cameras = kapture.io.csv.records_camera_from_file(records_camera_filepath)
    # 2: load trajectories
    trajectories_filepath = kapture.io.csv.get_csv_fullpath(kapture.Trajectories, kapture_dirpath)
    logger.debug(f'loading {trajectories_filepath}')
    trajectories = kapture.io.csv.trajectories_from_file(trajectories_filepath)
    # 3: find (timestamp, camera_id) that are both in records and trajectories.
    valid_keys = set(records_cameras.key_pairs()).intersection(set(trajectories.key_pairs()))
    # collect data for those timestamps.
    if keep_full_file_name:
        image_poses = ((k[1], records_cameras[k], trajectories[k]) for k in valid_keys)
    else:
        image_poses = ((k[1], path.basename(records_cameras[k]), trajectories[k]) for k in valid_keys)

    # prepend the camera name or drop it.
    if prepend_camera_name:
        image_poses = ((path.join(camera_id, image_filename), pose) for camera_id, image_filename, pose in image_poses)
    else:
        image_poses = ((image_filename, pose) for camera_id, image_filename, pose in image_poses)

    # write the files
    image_poses = {image_filename: pose
                   for image_filename, pose in image_poses}
    p = pathlib.Path(ltvl_filepath)
    os.makedirs(str(p.parent.resolve()), exist_ok=True)
    with open(ltvl_filepath, 'wt') as f:
        for image_filename, pose in tqdm(image_poses.items(), disable=logger.getEffectiveLevel() > logging.INFO):
            line = [image_filename] + pose.r_raw + pose.t_raw
            line = ' '.join(str(v) for v in line) + '\n'
            f.write(line)


def export_ltvl2020_command_line() -> None:
    """
    Do the LTVL 2020 export using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='convert file to Long-term Visual Localization challenge format '
                    '(https://www.visuallocalization.net/submission/).')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture directory')
    parser.add_argument('-o', '--output', required=True, help='output file.')
    parser.add_argument('-p', '--prepend_cam', action='store_true', default=False,
                        help='prepend camera names to filename (required for some dataset).')
    parser.add_argument('--full_file_name', action='store_true', default=False,
                        help='keep the full file name (default: seq-0/image1.png -> image1.png).')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    export_ltvl(args.input, args.output, args.prepend_cam, args.full_file_name)


if __name__ == '__main__':
    export_ltvl2020_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a UTBM dataset stored as rosbags: one with images from several camera,
and one with GPS position associated to the images recorded.

This is an outdoor dataset recorded with a car in various conditions.
More info at:
https://epan-utbm.github.io/utbm_robocar_dataset/
"""

import argparse
import logging
import sys
# kapture
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture.core.Sensors
import kapture.utils.logging

# rosbag
try:
    import rosbag  # noqa: F401
    from kapture.converter.ros_tools.import_rosbag import RosBagImporter
    has_rosbag = True
except ModuleNotFoundError:
    has_rosbag = False

from kapture.converter.ros_tools.import_utbm_sensor import BB2_CAMERA_IDENTIFIERS, TOPICS_BB2
from kapture.converter.ros_tools.import_utbm_sensor import XB3_CAMERA_IDENTIFIERS, TOPICS_XB3
from kapture.converter.ros_tools.import_utbm_sensor import import_utbm_sensors

logger = logging.getLogger('rosbag')  # Using global logger


def import_utbm_rosbag(args) -> None:
    """
    Import the UTBM rosbag with the parameters.

    :param args: arguments to use
    """
    # Read camera calibration files to make a sensors object
    sensors = import_utbm_sensors(args.camera_info)
    importer = RosBagImporter(args.bag_file, None, sensors, args.kapture_output, args.force)
    if args.bb2:
        image_topics = TOPICS_BB2
        camera_identifiers = BB2_CAMERA_IDENTIFIERS
    elif args.xb3:
        image_topics = TOPICS_XB3
        camera_identifiers = XB3_CAMERA_IDENTIFIERS
    else:
        image_topics = args.image_topics
        if not image_topics:
            image_topics = TOPICS_BB2
        camera_identifiers = args.camera_identifiers
        if not camera_identifiers:
            camera_identifiers = BB2_CAMERA_IDENTIFIERS
    importer.import_multi_camera(None, image_topics, camera_identifiers, False, False, args.percent)
    importer.save_to_kapture()


def import_utbm_rosbag_command_line() -> None:
    """
    Do the UTBM to kapture import using the command line parameters provided by the user.
    """
    parser = argparse.ArgumentParser(description='Imports UTBM rosbags to kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const', dest='verbose',
                                  const=logging.CRITICAL)
    parser.add_argument('-y', '--force', action='store_true', default=False,
                        help='silently delete kapture data if already exists.')
    # create the parser for the import command #########################################################################
    parser.add_argument('-b', '--bag_file', default='input.bag',
                        help='bag file path', required=False)
    parser.add_argument('--camera_info', nargs="*",
                        help='list of cameras calibration yaml files as published on UTBM github', required=True)
    parser.add_argument('-i', '--image_topics', nargs="*",
                        help=f'list of image topics to import, defaults to BB2 values {TOPICS_BB2}',
                        required=False)
    parser.add_argument('-c', '--camera_identifiers', nargs="*",
                        help=f'list of camera identifiers as defined in the sensors file of the kapture rig'
                             ' and corresponding to the image topics in the same order;'
                             f' defaults to BB2 values {BB2_CAMERA_IDENTIFIERS}',
                        required=False)
    parser.add_argument('--bb2', action='store_true', default=False,
                        help='Set BB2 image topics and camera identifiers')
    parser.add_argument('--xb3', action='store_true', default=False,
                        help='Set XB3 image topics and camera identifiers')
    parser.add_argument('-p', '--percent', type=int, default=100,
                        help='percentage of images to keep', required=False)
    parser.add_argument('-k', '--kapture_output', required=True,
                        help='directory where to save Kapture files.')
    ####################################################################################################################
    args = parser.parse_args()

    # Change the default logging
    root_logger = logging.getLogger()
    root_logger.handlers = []
    ch = logging.StreamHandler()
    ch.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s::%(name)s: %(message)s'))
    logger.addHandler(ch)
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    import_utbm_rosbag(args)


if __name__ == '__main__':
    if not has_rosbag:
        logger.fatal("You don't have ROS or import_rosbag installed")
        sys.exit(-1)
    try:
        import_utbm_rosbag_command_line()
    except (AssertionError, OSError, ValueError) as any_ex:
        logger.fatal(f'Fatal error: {any_ex}')



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Export the list of images of a kapture into a file listing the image files, and optionally their camera parameters.
"""

from kapture.core.Records import RecordBluetooth, RecordGnss, RecordGyroscope,\
    RecordsAccelerometer, RecordsDepth, RecordsMagnetic
import logging
import os
import argparse
# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.csv import kapture_from_dir
from kapture.utils.paths import safe_remove_file


logger = logging.getLogger('export_image_list')


def export_image_list(kapture_path: str, output_path: str, export_camera_params: bool, force: bool) -> None:
    """
    Export image list in a text file.

    :param kapture_path: top directory of the kapture
    :param output_path: path of the image list file
    :param export_camera_params: if True, add camera parameters after every file name
    :param force: Silently overwrite image list file if already exists.
    """
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    safe_remove_file(output_path, force)
    skip_heavy_useless = [kapture.Trajectories,
                          kapture.RecordsLidar, kapture.RecordsWifi,
                          kapture.RecordsAccelerometer, kapture.RecordBluetooth,
                          kapture.RecordGnss, kapture.RecordGyroscope,
                          kapture.RecordsDepth, kapture.RecordsMagnetic,
                          kapture.Keypoints, kapture.Descriptors, kapture.GlobalFeatures,
                          kapture.Matches, kapture.Points3d, kapture.Observations]
    kapture_to_export = kapture_from_dir(kapture_path, skip_list=skip_heavy_useless)
    output_content = []
    logger.info('starting conversion...')
    for _, sensor_id, filename in kapture.flatten(kapture_to_export.records_camera, is_sorted=True):
        line = filename
        if export_camera_params:
            camera = kapture_to_export.sensors[sensor_id]
            assert isinstance(camera, kapture.Camera)
            line += ' ' + ' '.join(camera.sensor_params)
        output_content.append(line)

    logger.info('writing exported data...')
    with open(output_path, 'w') as fid:
        fid.write('\n'.join(output_content))


def export_image_list_command_line() -> None:
    """
    Do the image list export using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='export kapture data image list to text file.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '--force', action='store_true', default=False,
                        help='force delete output if already exists.')
    # export ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to kapture root folder')
    parser.add_argument('-o', '--output', required=True, help='output path to text file')
    parser.add_argument('-c', '--camera-params', action='store_true', default=False,
                        help='also export camera params in the text file')
    ####################################################################################################################
    args = parser.parse_args()

    """
    image_name camera_type camera_params
    example: image_name SIMPLE_RADIAL w h f cx cy r
    or
    image_name
    """

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    export_image_list(args.input, args.output, args.camera_params, args.force)


if __name__ == '__main__':
    export_image_list_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Merge kapture objects with remapping of identifiers.
"""

from kapture.io.tar import TarCollection
from typing import List, Optional, Type, Dict

import kapture
from kapture.io.records import TransferAction, get_image_fullpath
from kapture.utils.Collections import get_new_if_not_empty

from .merge_reconstruction import merge_keypoints_collections, merge_descriptors_collections
from .merge_reconstruction import merge_global_features_collections, merge_matches_collections
from .merge_reconstruction import merge_points3d_and_observations, merge_points3d
from .merge_records_data import merge_records_data


def get_sensors_mapping(sensors: kapture.Sensors, offset: int = 0) -> Dict[str, str]:
    """
    Creates list of sensor names,identifiers

    :param sensors: list of sensor definitions
    :param offset: optional offset for the identifier numbers
    :return: mapping of sensor names to identifiers
    """
    return {k: f'sensor{v}' for k, v in zip(sensors.keys(), range(offset, offset + len(sensors)))}


def get_rigs_mapping(rigs: kapture.Rigs, offset: int = 0) -> Dict[str, str]:
    """
    Creates list of rig names,identifiers

    :param rigs: list of rig definitions
    :param offset: optional offset for the identifier numbers
    :return: mapping of rig names to identifiers
    """
    return {k: f'rig{v}' for k, v in zip(rigs.keys(), range(offset, offset + len(rigs)))}


def merge_table_key1(
        table_list,
        sensor_mappings: List[Dict[str, str]],
        table_constructor,
):
    """
    Merge several table with 1 key (Only device_id) into one.
    If multiple entry for a key keep only the first one.

    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :return table_merged

    """
    assert len(table_list) > 0
    assert len(table_list) == len(sensor_mappings)
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table, sensor_mapping in zip(table_list, sensor_mappings):
        for sensor_id, entry in kapture.flatten(table):
            new_sensor_id = sensor_mapping[sensor_id]
            table_merged[new_sensor_id] = entry
    return table_merged


def merge_table_key2(
        table_list,
        sensor_mappings: List[Dict[str, str]],
        table_constructor,
):
    """
    Merge several table with 2 keys (eg. timestamps, device_id)  into one.
    If multiple entry for a key keep only the first one.

    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :return table_merged

    """
    assert len(table_list) > 0
    assert len(table_list) == len(sensor_mappings)
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table, sensor_mapping in zip(table_list, sensor_mappings):
        for key1, sensor_id, entry in kapture.flatten(table):
            new_sensor_id = sensor_mapping[sensor_id]
            table_merged[key1, new_sensor_id] = entry
    return table_merged


def merge_table_key3(
        table_list,
        sensor_mappings: List[Dict[str, str]],
        table_constructor,
        subdict_constructor=dict,
):
    """
    Merge several table with 2 keys (eg. timestamps, device_id)  into one.
    If multiple entry for a key keep only the first one.

    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :param table_list: list of table to merge.
    :param table_constructor: the class type of table.
    :param subdict_constructor: used to create a new Dict type
    :return table_merged

    """
    assert len(table_list) > 0
    assert len(table_list) == len(sensor_mappings)
    table_list = [table for table in table_list if table is not None]
    if not all(isinstance(table, table_constructor) for table in table_list):
        raise TypeError('unexpected type.')
    table_merged = table_constructor()
    for table, sensor_mapping in zip(table_list, sensor_mappings):
        for key1, sensor_id, key3, entry in kapture.flatten(table):
            new_sensor_id = sensor_mapping[sensor_id]
            if (key1, new_sensor_id) not in table_merged:
                # if timestamp, sensor_id not there yet, create an instance of dict record
                table_merged[key1, new_sensor_id] = subdict_constructor()
            table_merged[key1, new_sensor_id].setdefault(key3, entry)
    return table_merged


def merge_sensors(
        sensors_list: List[Optional[kapture.Sensors]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.Sensors:
    """
    Merge several sensors list into one list with new identifiers.

    :param sensors_list: list of sensors definitions to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged sensors definitions
    """
    return merge_table_key1(
        table_list=sensors_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.Sensors
    )


def merge_rigs(
        rigs_list: List[Optional[kapture.Rigs]],
        rig_mappings: List[Dict[str, str]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.Rigs:
    """
    Merge several rigs list into one list with new identifiers for the rigs and the sensors.

    :param rigs_list: list of rigs definitions to merge
    :param rig_mappings: mapping of the rigs identifiers to their new identifiers
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged rigs definitions
    """
    assert len(rigs_list) > 0
    assert len(rigs_list) == len(rig_mappings)
    assert len(rigs_list) == len(sensor_mappings)

    merged_rigs = kapture.Rigs()
    for rigs, rig_mapping, sensor_mapping in zip(rigs_list, rig_mappings, sensor_mappings):
        if rigs is None:
            continue
        for rig_id, sensor_id in rigs.key_pairs():
            new_rig_id = rig_mapping[rig_id]
            new_sensor_id = sensor_mapping[sensor_id]
            merged_rigs[(new_rig_id, new_sensor_id)] = rigs[(rig_id, sensor_id)]
    return merged_rigs


def merge_trajectories(
        trajectories_list: List[Optional[kapture.Trajectories]],
        rig_mappings: List[Dict[str, str]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.Trajectories:
    """
    Merge several trajectories list into one list with new identifiers for the rigs and the sensors.

    :param trajectories_list: list of trajectories to merge
    :param rig_mappings: mapping of the rigs identifiers to their new identifiers
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged trajectories
    """
    assert len(trajectories_list) > 0
    assert len(trajectories_list) == len(rig_mappings)
    assert len(trajectories_list) == len(sensor_mappings)

    merged_trajectories = kapture.Trajectories()
    for trajectories, rig_mapping, sensor_mapping in zip(trajectories_list, rig_mappings, sensor_mappings):
        if trajectories is None:
            continue
        for timestamp, sensor_id, pose in kapture.flatten(trajectories):
            if sensor_id in rig_mapping:
                new_sensor_id = rig_mapping[sensor_id]
            else:
                new_sensor_id = sensor_mapping[sensor_id]
            merged_trajectories[(timestamp, new_sensor_id)] = pose
    return merged_trajectories


def merge_records_camera(
        records_camera_list: List[Optional[kapture.RecordsCamera]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsCamera:
    """
    Merge several camera records list into one list with new identifiers for the sensors.

    :param records_camera_list: list of camera records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged camera records
    """
    return merge_table_key2(
        table_list=records_camera_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsCamera
    )


def merge_records_lidar(
        records_lidar_list: List[Optional[kapture.RecordsLidar]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsLidar:
    """
    Merge several lidar records list into one list with new identifiers for the sensors.

    :param records_lidar_list: list of lidar records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged lidar records
    """
    return merge_table_key2(
        table_list=records_lidar_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsLidar
    )


def merge_records_wifi(
        records_wifi_list: List[Optional[kapture.RecordsWifi]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsWifi:
    """
    Merge several wifi records list into one list with new identifiers for the sensors.

    :param records_wifi_list: list of wifi records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged wifi records
    """
    return merge_table_key3(
        table_list=records_wifi_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsWifi,
        subdict_constructor=kapture.RecordsWifi.record_type
    )


def merge_records_bluetooth(
        records_bluetooth_list: List[Optional[kapture.RecordsBluetooth]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsBluetooth:
    """
    Merge several bluetooth records list into one list with new identifiers for the sensors.

    :param records_bluetooth_list: list of wifi records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged bluetooth records
    """
    return merge_table_key3(
        table_list=records_bluetooth_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsBluetooth,
        subdict_constructor=kapture.RecordsBluetooth.record_type
    )


def merge_records_gnss(
        records_gnss_list: List[Optional[kapture.RecordsGnss]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsGnss:
    """
    Merge several gnss records list into one list with new identifiers for the sensors.

    :param records_gnss_list: list of gnss records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged gnss records
    """
    return merge_table_key2(
        table_list=records_gnss_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsGnss
    )


def merge_records_accelerometer(
        records_accelerometer_list: List[Optional[kapture.RecordsAccelerometer]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsAccelerometer:
    """
    Merge several accelerometer records list into one list with new identifiers for the sensors.

    :param records_accelerometer_list: list of accelerometer records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged accelerometer records
    """
    return merge_table_key2(
        table_list=records_accelerometer_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsAccelerometer
    )


def merge_records_gyroscope(
        records_gyroscope_list: List[Optional[kapture.RecordsGyroscope]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsGyroscope:
    """
    Merge several gyroscope records list into one list with new identifiers for the sensors.

    :param records_gyroscope_list: list of gyroscope records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged gyroscope records
    """
    return merge_table_key2(
        table_list=records_gyroscope_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsGyroscope
    )


def merge_records_magnetic(
        records_magnetic_list: List[Optional[kapture.RecordsMagnetic]],
        sensor_mappings: List[Dict[str, str]]) -> kapture.RecordsMagnetic:
    """
    Merge several magnetic records list into one list with new identifiers for the sensors.

    :param records_magnetic_list: list of magnetic records to merge
    :param sensor_mappings: mapping of the sensor identifiers to their new identifiers
    :return: merged magnetic records
    """
    return merge_table_key2(
        table_list=records_magnetic_list,
        sensor_mappings=sensor_mappings,
        table_constructor=kapture.RecordsMagnetic
    )


def merge_remap(kapture_list: List[kapture.Kapture],  # noqa: C901: function a bit long but not too complex
                skip_list: List[Type],
                data_paths: List[str],
                tarcollection_list: List[TarCollection],
                kapture_path: str,
                images_import_method: TransferAction) -> kapture.Kapture:
    """
    Merge multiple kapture while keeping ids (sensor_id) identical in merged and inputs.

    :param kapture_list: list of kapture to merge.
    :param skip_list: input optional types to not merge. sensors and rigs are unskippable
    :param data_paths: list of path to root path directory in same order as mentioned in kapture_list.
    :param tarcollection_list: list of opened tar archives same order as mentioned in kapture_list.
    :param kapture_path: directory root path to the merged kapture.
    :param images_import_method: method to transfer image files
    :return: merged kapture object
    """
    merged_kapture = kapture.Kapture()

    # find new sensor ids / rig ids
    sensors_mapping = []
    rigs_mapping = []
    _compute_new_ids(kapture_list, rigs_mapping, sensors_mapping)

    # concatenate all sensors with the remapped ids
    new_sensors = merge_sensors([a_kapture.sensors for a_kapture in kapture_list], sensors_mapping)
    # if merge_sensors returned an empty object, keep merged_kapture.sensors to None
    merged_kapture.sensors = get_new_if_not_empty(new_sensors, merged_kapture.sensors)

    # concatenate all rigs with the remapped ids
    new_rigs = merge_rigs([a_kapture.rigs for a_kapture in kapture_list], rigs_mapping, sensors_mapping)
    # if merge_rigs returned an empty object, keep merged_kapture.rigs to None
    merged_kapture.rigs = get_new_if_not_empty(new_rigs, merged_kapture.rigs)

    # all fields below can be skipped with skip_list
    # we do not assign the properties when the merge evaluate to false, we keep it as None
    if kapture.Trajectories not in skip_list:
        new_trajectories = merge_trajectories([a_kapture.trajectories for a_kapture in kapture_list],
                                              rigs_mapping,
                                              sensors_mapping)
        merged_kapture.trajectories = get_new_if_not_empty(new_trajectories, merged_kapture.trajectories)

    if kapture.RecordsCamera not in skip_list:
        new_records_camera = merge_records_camera([a_kapture.records_camera for a_kapture in kapture_list],
                                                  sensors_mapping)
        merged_kapture.records_camera = get_new_if_not_empty(new_records_camera, merged_kapture.records_camera)

        merge_records_data([[image_name
                             for _, _, image_name in kapture.flatten(every_kapture.records_camera)]
                            if every_kapture.records_camera is not None else []
                            for every_kapture in kapture_list],
                           [get_image_fullpath(data_path, image_filename=None) for data_path in data_paths],
                           kapture_path,
                           images_import_method)
    if kapture.RecordsLidar not in skip_list:
        new_records_lidar = merge_records_lidar([a_kapture.records_lidar for a_kapture in kapture_list],
                                                sensors_mapping)
        merged_kapture.records_lidar = get_new_if_not_empty(new_records_lidar, merged_kapture.records_lidar)
    if kapture.RecordsWifi not in skip_list:
        new_records_wifi = merge_records_wifi([a_kapture.records_wifi for a_kapture in kapture_list],
                                              sensors_mapping)
        merged_kapture.records_wifi = get_new_if_not_empty(new_records_wifi, merged_kapture.records_wifi)
    if kapture.RecordsBluetooth not in skip_list:
        new_records_bluetooth = merge_records_bluetooth([a_kapture.records_bluetooth for a_kapture in kapture_list],
                                                        sensors_mapping)
        merged_kapture.records_bluetooth = get_new_if_not_empty(new_records_bluetooth, merged_kapture.records_bluetooth)
    if kapture.RecordsGnss not in skip_list:
        new_records_gnss = merge_records_gnss([a_kapture.records_gnss for a_kapture in kapture_list],
                                              sensors_mapping)
        merged_kapture.records_gnss = get_new_if_not_empty(new_records_gnss,
                                                           merged_kapture.records_gnss)
    if kapture.RecordsAccelerometer not in skip_list:
        new_records_accelerometer = merge_records_accelerometer(
            [a_kapture.records_accelerometer for a_kapture in kapture_list],
            sensors_mapping)
        merged_kapture.records_accelerometer = get_new_if_not_empty(new_records_accelerometer,
                                                                    merged_kapture.records_accelerometer)
    if kapture.RecordsGyroscope not in skip_list:
        new_records_gyroscope = merge_records_gyroscope(
            [a_kapture.records_gyroscope for a_kapture in kapture_list],
            sensors_mapping)
        merged_kapture.records_gyroscope = get_new_if_not_empty(new_records_gyroscope,
                                                                merged_kapture.records_gyroscope)
    if kapture.RecordsMagnetic not in skip_list:
        new_records_magnetic = merge_records_magnetic(
            [a_kapture.records_magnetic for a_kapture in kapture_list],
            sensors_mapping)
        merged_kapture.records_magnetic = get_new_if_not_empty(new_records_magnetic,
                                                               merged_kapture.records_magnetic)

    # for the reconstruction, except points and observations, the files are copied with shutil.copy
    # if kapture_path evaluates to False, all copies will be skipped (but classes will be filled normally)
    if kapture.Keypoints not in skip_list:
        keypoints = [a_kapture.keypoints for a_kapture in kapture_list]
        keypoints_not_none = [k for k in keypoints if k is not None]
        if len(keypoints_not_none) > 0:
            new_keypoints = merge_keypoints_collections(keypoints, data_paths, kapture_path, tarcollection_list)
            merged_kapture.keypoints = get_new_if_not_empty(new_keypoints, merged_kapture.keypoints)
    if kapture.Descriptors not in skip_list:
        descriptors = [a_kapture.descriptors for a_kapture in kapture_list]
        descriptors_not_none = [k for k in descriptors if k is not None]
        if len(descriptors_not_none) > 0:
            new_descriptors = merge_descriptors_collections(descriptors, data_paths, kapture_path, tarcollection_list)
            merged_kapture.descriptors = get_new_if_not_empty(new_descriptors, merged_kapture.descriptors)
    if kapture.GlobalFeatures not in skip_list:
        global_features = [a_kapture.global_features for a_kapture in kapture_list]
        global_features_not_none = [k for k in global_features if k is not None]
        if len(global_features_not_none) > 0:
            new_global_features = merge_global_features_collections(global_features, data_paths,
                                                                    kapture_path, tarcollection_list)
            merged_kapture.global_features = get_new_if_not_empty(new_global_features, merged_kapture.global_features)
    if kapture.Matches not in skip_list:
        matches = [a_kapture.matches for a_kapture in kapture_list]
        new_matches = merge_matches_collections(matches, data_paths, kapture_path, tarcollection_list)
        merged_kapture.matches = get_new_if_not_empty(new_matches, merged_kapture.matches)

    if kapture.Points3d not in skip_list and kapture.Observations not in skip_list:
        points_and_obs = [(a_kapture.points3d, a_kapture.observations) for a_kapture in kapture_list]
        new_points, new_observations = merge_points3d_and_observations(points_and_obs)
        merged_kapture.points3d = get_new_if_not_empty(new_points, merged_kapture.points3d)
        merged_kapture.observations = get_new_if_not_empty(new_observations, merged_kapture.observations)
    elif kapture.Points3d not in skip_list:
        points = [a_kapture.points3d for a_kapture in kapture_list]
        new_points = merge_points3d(points)
        merged_kapture.points3d = get_new_if_not_empty(new_points, merged_kapture.points3d)
    return merged_kapture


def _compute_new_ids(kapture_list, rigs_mapping, sensors_mapping):
    sensor_offset = 0
    rigs_offset = 0
    for every_kapture in kapture_list:
        if every_kapture.sensors is not None:
            sensors_mapping.append(get_sensors_mapping(every_kapture.sensors, sensor_offset))
            sensor_offset += len(every_kapture.sensors)
        else:
            sensors_mapping.append({})

        if every_kapture.rigs is not None:
            rigs_mapping.append(get_rigs_mapping(every_kapture.rigs, rigs_offset))
            rigs_offset += len(every_kapture.rigs)
        else:
            rigs_mapping.append({})



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Represents a 6 dimensions pose.
"""

import math
import numpy as np
import quaternion
from typing import List, Optional
from numba import njit


class PoseTransform:
    """
    Brief: Transforms points from reference frame to another.
            In case of
             - Trajectories: transforms points from world to device,
             - Rig: transforms points from rig to camera.
    """

    def __init__(self, r=[1, 0, 0, 0], t=[0, 0, 0]):
        """
        Creates a Pose transform (eg. world to device) with given rotation + translation

        :param r: rotation part expressed as a quaternion (a list of floats, a numpy array, or a numpy quaternion)
        :param t: translation part expressed as a vector ((a list of floats, a numpy array)
        """

        # quaternion
        if isinstance(r, quaternion.quaternion):
            # already a quaternion : copy aziz (lumière)
            self._r = r
        elif isinstance(r, (list, np.ndarray, np.generic)):
            # its a list or numpy => convert to quaternion if valid
            self._r = quaternion.from_float_array(r)
        else:  # unknown entry
            self._r = None

        if self._r is not None and np.isnan(np.sum(self._r)):
            # check no nan in the array (meaning failure)
            # sum trick from https://stackoverflow.com/questions/6736590/fast-check-for-nan-in-numpy
            self._r = None

        # translation
        if isinstance(t, (np.ndarray, np.generic)):
            #  numpy : copy aziz lumière
            self._t = t
        elif isinstance(t, list):
            # its a list: convert to numpy if valid
            self._t = np.array(t, dtype=np.float)
        else:
            self._t = None

        if self._t is not None:
            assert isinstance(self._t, (np.ndarray, np.generic))
            if np.isnan(np.sum(self._t)):
                # check no nan in the array (meaning failure)
                # sum trick from https://stackoverflow.com/questions/6736590/fast-check-for-nan-in-numpy
                self._t = None
            else:
                # make sure af the shape of t
                self._t = self._t.reshape((3, 1))

    @property
    def r(self):
        """
        :return: rotation as quaternion
        """
        return self._r

    @property
    def r_raw(self) -> Optional[List[float]]:
        """
        :return: rotation as list of float
        """
        return quaternion.as_float_array(self._r).tolist() if self._r is not None else None

    @property
    def t(self):
        """
        :return: translation as numpy ndarray
        """
        return self._t

    @property
    def t_raw(self) -> Optional[List[float]]:
        """
        :return: translation as list of float
        """
        return self._t.flatten().tolist() if self._t is not None else None

    def inverse(self) -> 'PoseTransform':
        """
        :return: inverted pose
        """
        assert self._r is not None
        assert self._t is not None
        r_inv = self.r.inverse()
        rotation_inv_matrix = np.empty((3, 3), dtype=np.float)
        rotation_inv_as_np = np.array([r_inv.w, r_inv.x, r_inv.y, r_inv.z])
        _as_rotation_matrix_njit(rotation_inv_as_np, rotation_inv_matrix)
        t_inv = np.matmul(rotation_inv_matrix, self.t * -1.0)
        pose_inv = PoseTransform.__new__(PoseTransform)
        pose_inv._r = r_inv
        pose_inv._t = t_inv
        return pose_inv

    @staticmethod
    def compose(pose_list: List['PoseTransform']) -> 'PoseTransform':
        """
        Merges multiple pose transformation into a single one.

        :param pose_list: the list of pose transformation to be merged. They are merged from right to left.
        :return: the pose transformation being the composition of the given ones.
        """
        assert isinstance(pose_list, list)
        pose_composed = pose_list[0]
        for pose_current in pose_list[1:]:
            # shrink the poses from the left
            # assert pose_current._r is not None
            # assert pose_current._t is not None
            new_r = pose_composed.r * pose_current.r
            rotation_matrix = np.empty((3, 3), dtype=np.float)
            rotation_composed_as_np = np.array([pose_composed.r.w, pose_composed.r.x,
                                                pose_composed.r.y, pose_composed.r.z])
            _as_rotation_matrix_njit(rotation_composed_as_np, rotation_matrix)
            new_t = np.add(np.matmul(rotation_matrix, pose_current.t), pose_composed.t)

            pose_composed = PoseTransform.__new__(PoseTransform)
            pose_composed._r = new_r
            pose_composed._t = new_t
        return pose_composed

    def transform_points(self, points3d: np.ndarray) -> np.ndarray:
        """
        Applies the self transformation to the given 3d points.

        :param points3d: input 3d points, stored row wise (one point per row)
        :return: another array with the 3d points transformed using the PoseTransform.
        """
        assert self._r is not None
        assert self._t is not None
        assert isinstance(points3d, np.ndarray)

        if points3d.shape[1] == 6:  # expunge RGB
            points3d = points3d[:, 0:3]
        points3d = points3d.transpose()
        rotation_matrix = np.empty((3, 3), dtype=np.float)
        rotation_as_np = np.array([self.r.w, self.r.x, self.r.y, self.r.z])
        _as_rotation_matrix_njit(rotation_as_np, rotation_matrix)
        points3d = np.add(np.matmul(rotation_matrix, points3d), self.t)
        return points3d.transpose()

    def __repr__(self) -> str:
        return 'r:{},  t:{}'.format(self.r_raw, self.t_raw)

    def __eq__(self, other) -> bool:
        if (self._t is None and other.t is not None) or (self._t is not None and other.t is None):
            return False
        if (self._r is None and other.r is not None) or (self._r is not None and other.r is None):
            return False
        self_trans = self.t_raw
        other_trans = other.t_raw
        for left, right in zip(self_trans, other_trans):
            # Check distances at 10-2 millimeters
            if not math.isclose(left, right, rel_tol=1.e-05, abs_tol=1.e-05):
                return False
        self_rot = self.r_raw
        other_rot = other.r_raw
        for left, right in zip(self_rot, other_rot):
            # Check angles at 10-2 degrees
            if not math.isclose(left, right, rel_tol=1.e-02, abs_tol=1.e-02):
                return False
        return True


# https://github.com/moble/quaternion/blob/main/src/quaternion/__init__.py#L200
@njit
def _as_rotation_matrix_njit(q, rot_mat):
    q_norm = q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]
    if abs(q_norm-1.0) < 1e-14:
        rot_mat[0, 0] = 1 - 2*(q[2]**2 + q[3]**2)
        rot_mat[0, 1] = 2*(q[1]*q[2] - q[3]*q[0])
        rot_mat[0, 2] = 2*(q[1]*q[3] + q[2]*q[0])
        rot_mat[1, 0] = 2*(q[1]*q[2] + q[3]*q[0])
        rot_mat[1, 1] = 1 - 2*(q[1]**2 + q[3]**2)
        rot_mat[1, 2] = 2*(q[2]*q[3] - q[1]*q[0])
        rot_mat[2, 0] = 2*(q[1]*q[3] - q[2]*q[0])
        rot_mat[2, 1] = 2*(q[2]*q[3] + q[1]*q[0])
        rot_mat[2, 2] = 1 - 2*(q[1]**2 + q[2]**2)
    else:
        rot_mat[0, 0] = 1 - 2*(q[2]**2 + q[3]**2)/q_norm
        rot_mat[0, 1] = 2*(q[1]*q[2] - q[3]*q[0])/q_norm
        rot_mat[0, 2] = 2*(q[1]*q[3] + q[2]*q[0])/q_norm
        rot_mat[1, 0] = 2*(q[1]*q[2] + q[3]*q[0])/q_norm
        rot_mat[1, 1] = 1 - 2*(q[1]**2 + q[3]**2)/q_norm
        rot_mat[1, 2] = 2*(q[2]*q[3] - q[1]*q[0])/q_norm
        rot_mat[2, 0] = 2*(q[1]*q[3] - q[2]*q[0])/q_norm
        rot_mat[2, 1] = 2*(q[2]*q[3] + q[1]*q[0])/q_norm
        rot_mat[2, 2] = 1 - 2*(q[1]**2 + q[2]**2)/q_norm
    return



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Additional collections and base python objects.
"""

import collections
from enum import Enum
from typing import Any, Optional


class AutoEnum(Enum):
    """
    When the auto operator is used, value = name
    see https://docs.python.org/3/library/enum.html#using-automatic-values
    """
    def _generate_next_value_(name, start, count, last_values):
        return name


class LimitedDictionary(collections.OrderedDict):
    """
    An ordered dictionary with a limited size, evicting the oldest inserted when full
    """

    def __init__(self, maxsize=64, *args, **kwds):
        self._maxsize = maxsize
        super().__init__(*args, **kwds)

    def __getitem__(self, key):
        value = super().__getitem__(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        if len(self) > self._maxsize:
            oldest = next(iter(self))
            del self[oldest]


def get_new_if_not_empty(new_value: Any, old_value: Any) -> Any:
    """
    Return the new value if it is not None
    """
    return new_value if new_value else old_value


def try_get_only_key_from_collection(collection: Optional[dict]) -> Optional[str]:
    """
    3 scenarios:
    1 - dict is None or empty -> return None
    2 - dict has exactly one item, return its key
    3 - dict has more than one item, raise exception
    """
    if collection is None or len(collection) == 0:
        return None
    if len(collection) == 1:
        return next(iter(collection.keys()))
    raise ValueError(f'{collection} has more than 1 item, cannot decide which to pick')



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to print statistics about kapture data.
"""

import argparse
import logging
import sys
import contextlib
import os.path as path
from typing import Optional, Tuple
from datetime import datetime
import time

import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
import kapture.io.csv

logger = logging.getLogger('kapture_print')

# Consider as valid timestamps only if between these two dates
LOWER_RECORD_DATE_TIMESTAMP = time.mktime(datetime(year=1980, month=1, day=1, hour=0, minute=0, second=0).timetuple())
UPPER_RECORD_DATE_TIMESTAMP = time.mktime(datetime(year=2100, month=1, day=1, hour=0, minute=0, second=0).timetuple())


@contextlib.contextmanager
def open_or_stdout(filename=None):
    """
    Get the output to print into
    """
    # from https://stackoverflow.com/questions/17602878/how-to-handle-both-with-open-and-sys-stdout-nicely
    if filename and filename != '-':
        fh = open(filename, 'w')
    else:
        fh = sys.stdout

    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


def print_key_value(key, value, file, show_none=False) -> None:
    """
    Prints a key and its value
    """
    if value is not None or show_none:
        print(f'{key:25}: {value}', file=file)


def print_title(title, file) -> None:
    """
    Prints the title
    """
    print(f'=== {title:^25} ===', file=file)


def print_sensors(kapture_data, output_stream, show_detail, show_all) -> None:
    """
    Prints the sensors (only) to the output stream
    """
    if not show_detail:
        print_key_value('nb sensors', len(kapture_data.sensors), file=output_stream, show_none=show_all)
    else:
        print_title('sensors', file=output_stream)
        cam_ids = [s_id for s_id, sensor in kapture_data.sensors.items()
                   if sensor.sensor_type in kapture.ALL_CAMERA_SENSOR_TYPES]
        print_key_value(' ├─ nb cameras', len(cam_ids), file=output_stream, show_none=show_all)
        lidar_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'lidar']
        print_key_value(' ├─ nb lidar', len(lidar_ids), file=output_stream, show_none=show_all)
        wifi_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'wifi']
        print_key_value(' ├─ nb wifi', len(wifi_ids), file=output_stream, show_none=show_all)
        bluetooth_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'bluetooth']
        print_key_value(' ├─ nb bluetooth', len(bluetooth_ids), file=output_stream, show_none=show_all)
        gnss_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'gnss']
        print_key_value(' ├─ nb gnss', len(gnss_ids), file=output_stream, show_none=show_all)
        accelerometer_ids = [s_id for s_id, sensor in kapture_data.sensors.items()
                             if sensor.sensor_type == 'accelerometer']
        print_key_value(' ├─ nb accelerometer', len(accelerometer_ids), file=output_stream, show_none=show_all)
        gyroscope_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'gyroscope']
        print_key_value(' ├─ nb gyroscope', len(gyroscope_ids), file=output_stream, show_none=show_all)
        magnetic_ids = [s_id for s_id, sensor in kapture_data.sensors.items() if sensor.sensor_type == 'magnetic']
        print_key_value(' ├─ nb magnetic', len(magnetic_ids), file=output_stream, show_none=show_all)
        print_key_value(' └─ nb sensors total', len(kapture_data.sensors), file=output_stream, show_none=show_all)


def print_rigs(kapture_data, output_stream, show_detail, show_all) -> None:
    """
    Prints the rigs to the output stream
    """
    nb_rigs = len(kapture_data.rigs) if kapture_data.rigs is not None else None
    if not show_detail:
        print_key_value('nb rigs', nb_rigs, file=output_stream, show_none=show_all)
    else:
        print_title('rigs', file=output_stream)
        if kapture_data.rigs is not None:
            for rig_id, rig in kapture_data.rigs.items():
                s_ids = [s_id for s_id in kapture_data.rigs[rig_id]]
                print_key_value(' ├─ rig', rig_id, file=output_stream, show_none=show_all)
                for sensor_id in s_ids:
                    sensor_type = 'unknown'
                    if sensor_id in kapture_data.sensors:
                        sensor_type = kapture_data.sensors[sensor_id].sensor_type
                    if sensor_id in kapture_data.rigs:
                        sensor_type = 'rig'
                    print_key_value(f' │   ├─ {sensor_type}', sensor_id, file=output_stream, show_none=show_all)
                print_key_value(' │   └─ nb sensors total', len(s_ids), file=output_stream, show_none=show_all)
        print_key_value(' └─ nb rigs total', nb_rigs, file=output_stream, show_none=show_all)


FACTOR_TO_SECONDS = {
    'second': 1.0,
    'millisecond': 1.e-3,
    'microsecond': 1.e-6,
    'nanoseconds': 1.e-9
}


def guess_timestamp_posix_unit(timestamp: int) -> Optional[str]:
    """
    Guess the unit of a timestamp based on its value.

    :param timestamp: the timestamp value
    :return: 'posix' or 'posix-ms' or 'index'
    """
    assert isinstance(timestamp, int)

    for unit, factor in FACTOR_TO_SECONDS.items():
        if LOWER_RECORD_DATE_TIMESTAMP < factor * float(timestamp) < UPPER_RECORD_DATE_TIMESTAMP:
            return unit

    # none worked, its not posix
    return None


def format_timestamp_range(
        timestamp_range: Tuple[int, int],
        timestamp_unit: Optional[str],
        timestamp_format: Optional[str]
) -> str:
    """
    Format a time range with unit.

    :param timestamp_range: range of timestamps
    :param timestamp_unit: timestamp unit, can be: formatted_timestamp or auto
    :param timestamp_format: the timestamp format string
    :return string:
    """
    if timestamp_unit is None:
        timestamp_unit = 'index'
    elif timestamp_unit == 'auto':
        guessed_units = tuple(guess_timestamp_posix_unit(ts) for ts in timestamp_range)
        timestamp_unit = guessed_units[0] if guessed_units[0] == guessed_units[1] else 'index'

    if timestamp_unit in FACTOR_TO_SECONDS:
        try:
            factor = FACTOR_TO_SECONDS[timestamp_unit]
            dts = [datetime.utcfromtimestamp(ts * factor)
                   for ts in timestamp_range]
            # print the first timestamp completely
            if timestamp_format:
                return ' : '.join(dt.strftime(timestamp_format) for dt in dts)
            # if format not given, use ISO for 1st, and just print what change for second.
            timestamp_format_d = {
                'date': '%Y/%m/%d',
                'time': '%H:%M:%S.%f'
            }
            timestamp_parts = [
                {
                    pname: dt.strftime(pformat)
                    for pname, pformat in timestamp_format_d.items()
                }
                for dt in dts
            ]
            timestamp_str = timestamp_parts[0]['date']
            timestamp_str += ' '
            timestamp_str += timestamp_parts[0]['time']
            timestamp_str += '  :  '
            if timestamp_parts[0]['date'] != timestamp_parts[1]['date']:
                timestamp_str = timestamp_parts[0]['date']
                timestamp_str += ' '
            timestamp_str += timestamp_parts[1]['time']
            return timestamp_str
        except ValueError as _:  # noqa: F841
            return ' : '.join(str(ts) for ts in timestamp_range) + f' ** FAIL to parse as posix {timestamp_unit}'
    else:  # not posix
        return ' : '.join(str(ts) for ts in timestamp_range)


def print_records(
        kapture_data, output_stream, show_detail, show_all,
        timestamp_unit=None,
        timestamp_formatting=None
) -> None:
    """
    Prints the records and trajectories to the output stream
    """
    # records (+trajectories)
    for record_name in ['trajectories', 'records_camera', 'records_lidar', 'records_wifi', 'records_bluetooth',
                        'records_gnss', 'records_accelerometer', 'records_gyroscope', 'records_magnetic']:
        records_field = getattr(kapture_data, record_name)
        records = list(kapture.flatten(records_field))
        nb_record = None if records_field is None else len(records)
        if not show_detail:
            print_key_value(f'nb {record_name}', nb_record, output_stream, show_none=show_all)
        elif records_field is not None or show_all:
            print_title(f'{record_name}', output_stream)
            if records_field is not None and nb_record > 0:
                timestamp_range = (min(records_field), max(records_field))
                timestamp_range_str = format_timestamp_range(timestamp_range, timestamp_unit, timestamp_formatting)
                sensors_ids = records_field.sensors_ids
                print_key_value(' ├─ timestamp range', timestamp_range_str, output_stream, show_none=show_all)
                print_key_value(' ├─ sensors', f'{len(sensors_ids)}: {sensors_ids}', output_stream, show_none=show_all)
            print_key_value(' └─ nb total', nb_record, output_stream, show_none=show_all)


def print_features(kapture_data, output_stream, show_detail, show_all) -> None:
    """
    Prints the features to the output stream
    """
    # image features
    for feature_name in ['keypoints', 'descriptors', 'global_features']:
        feature_collection = getattr(kapture_data, feature_name)
        nb_feature_types = None if feature_collection is None else len(list(feature_collection.keys()))
        if not show_detail:
            print_key_value(f'nb types {feature_name}', nb_feature_types, file=output_stream, show_none=show_all)
            for feature_type, features in feature_collection.items():
                print_key_value(f' └─ nb images {feature_type} ', len(features), file=output_stream, show_none=show_all)
        elif feature_collection is not None or show_all:
            print_title(feature_name, file=output_stream)
            if feature_collection is not None:
                for feature_type, features in feature_collection.items():
                    print_key_value(' ├─ feature_type ', feature_type, file=output_stream, show_none=show_all)
                    print_key_value(' | ├─ kind ', features.type_name, file=output_stream, show_none=show_all)
                    print_key_value(' | ├─ data type', features.dtype.__name__, file=output_stream, show_none=show_all)
                    print_key_value(' | ├─ data size', features.dsize, file=output_stream, show_none=show_all)
                    if feature_name == 'descriptors':
                        print_key_value(' | ├─ keypoints_type', features.keypoints_type,
                                        file=output_stream, show_none=show_all)
                    if feature_name != 'keypoints':
                        print_key_value(' | ├─ metric_type', features.metric_type,
                                        file=output_stream, show_none=show_all)
                    print_key_value(' | └─ nb images ', len(list(features)), file=output_stream, show_none=show_all)
            print_key_value(f'nb types {feature_name}', nb_feature_types, file=output_stream, show_none=show_all)


def print_matches(kapture_data, output_stream, show_detail, show_all) -> None:
    """
    Prints the matches to the output stream
    """
    # matches
    nb_kpts_types = None if kapture_data.matches is None else len(list(kapture_data.matches.keys()))
    if not show_detail:
        print_key_value('nb types ', nb_kpts_types, output_stream, show_all)
        for kpt_type, matches in kapture_data.matches.items():
            print_key_value(f' └─ nb matching pairs {kpt_type} ', len(list(matches)),
                            output_stream, show_all)
    elif kapture_data.matches is not None or show_all:
        print_title('matches', output_stream)
        if kapture_data.matches is not None:
            for kpt_type, matches in kapture_data.matches.items():
                print_key_value(' ├─ keypoints_type ', kpt_type, output_stream, show_all)
                print_key_value(' | └─ nb pairs ', len(list(matches)), output_stream, show_all)
            print_key_value('nb types ', nb_kpts_types, output_stream, show_all)


def print_points(kapture_data, output_stream, show_detail, show_all) -> None:
    """
    Prints the 3D points (and observations) to the output stream
    """
    # points 3D
    nb_points3d = None if kapture_data.points3d is None else len(list(kapture_data.points3d))
    if not show_detail:
        print_key_value('nb points 3-D', nb_points3d, file=output_stream, show_none=show_all)
    elif kapture_data.points3d is not None or show_all:
        print_title('points 3-D', file=output_stream)
        print_key_value(' └─ nb points 3-D', nb_points3d, file=output_stream, show_none=show_all)
    # observations
    nb_observations_3d = len(kapture_data.observations) if kapture_data.observations is not None else None
    nb_observations_2d = len([feat
                              for observations in kapture_data.observations.values()
                              for feats in observations.values()
                              for feat in feats]) if kapture_data.observations is not None else None
    if not show_detail:
        print_key_value('nb observed 3-D points', nb_observations_3d, file=output_stream, show_none=show_all)
        print_key_value('nb observation 2-D points', nb_observations_2d, file=output_stream, show_none=show_all)
    elif kapture_data.observations is not None or show_all:
        print_title('Observations', file=output_stream)
        if kapture_data.observations is not None:
            print_key_value(' ├─ nb observed 3-D', nb_observations_3d, file=output_stream, show_none=show_all)
        print_key_value(' └─ nb observations 2-D', nb_observations_2d, file=output_stream, show_none=show_all)


def print_command_line() -> None:
    """
    Do the print using the parameters given on the command line.
    """

    parser = argparse.ArgumentParser(description='Print on stdout (or given file) statistics about kapture dataset.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--input', '-k', '--kapture', required=True,
                        help='path to kapture data root directory.')
    parser.add_argument('-o', '--output', required=False, default='-',
                        help='output file [stdout]')
    parser.add_argument('-a', '--all', action='store_true', default=False,
                        help='display all, even None')
    parser.add_argument('-d', '--detail', action='store_true', default=False,
                        help='display detailed')
    parser.add_argument('-t', '--timestamp_unit', choices=['auto', 'index', 'second', 'millisecond', 'microsecond'],
                        default='auto',
                        help='Force what timestamp really are (eg. millisecond) to display them nicely.'
                             'Auto means the program tries to guess. [auto]')
    parser.add_argument('-f', '--timestamp_formatting',
                        help='Tells what timestamp are, helps human display.')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    args.input = path.abspath(args.input)
    # load
    with kapture.io.csv.get_all_tar_handlers(args.input) as tar_handlers:
        kapture_data = kapture.io.csv.kapture_from_dir(args.input, tar_handlers=tar_handlers)
        do_print(kapture_data, args.input, args.output, args.detail, args.all,
                 args.timestamp_unit, args.timestamp_formatting)


def do_print(
        kapture_data: kapture,
        kapture_path: str,
        output_filepath: str,
        show_detail: bool,
        show_all: bool,
        timestamp_unit: str,
        timestamp_formatting: str,
) -> None:
    """
    Print out kapture data:

    :param kapture_data: input kapture data to print.
    :param kapture_path: full path to kapture directory.
    :param output_filepath: file path where to print. '-' means stdout.
    :param show_detail: If true, show details about data (in depth)
    :param show_all: If true, prints even if None
    :param timestamp_unit: tells the unit of timestamp (eg. posix)
    :param timestamp_formatting: how to format the timestamp on display
    """

    # print
    with open_or_stdout(output_filepath) as output_stream:
        if show_detail:
            print_title('general', file=output_stream)
            print_key_value('path', kapture_path, file=output_stream, show_none=show_all)
            print_key_value('version', kapture_data.format_version, file=output_stream, show_none=show_all)

        print_sensors(kapture_data, output_stream, show_detail, show_all)
        print_rigs(kapture_data, output_stream, show_detail, show_all)
        print_records(kapture_data, output_stream, show_detail, show_all, timestamp_unit, timestamp_formatting)
        print_features(kapture_data, output_stream, show_detail, show_all)
        print_matches(kapture_data, output_stream, show_detail, show_all)
        print_points(kapture_data, output_stream, show_detail, show_all)


if __name__ == '__main__':
    print_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Kapture to opensfm export functions.
"""

import os
import logging
import os.path as path
import numpy as np
import quaternion
import gzip
import pickle
import json
from tqdm import tqdm
from typing import Any, Dict, Optional
# kapture
import kapture
from kapture.io.csv import get_all_tar_handlers, kapture_from_dir
from kapture.utils.Collections import try_get_only_key_from_collection
from kapture.io.binary import TransferAction, transfer_files_from_dir
from kapture.io.features import get_keypoints_fullpath, image_keypoints_from_file
from kapture.io.features import get_descriptors_fullpath, image_descriptors_from_file
from kapture.io.features import get_matches_fullpath, image_matches_from_file
from kapture.io.records import get_record_fullpath


logger = logging.getLogger('opensfm')


"""
opensfm_project/
├── config.yaml
├── images/
├── masks/
├── gcp_list.txt
├── exif/
├── camera_models.json
├── features/
├── matches/
├── tracks.csv
├── reconstruction.json
├── reconstruction.meshed.json
└── undistorted/
    ├── images/
    ├── masks/
    ├── tracks.csv
    ├── reconstruction.json
    └── depthmaps/
        └── merged.ply
"""

"""
reconstruction.json: [RECONSTRUCTION, ...]

RECONSTRUCTION: {
    "cameras": {
        CAMERA_ID: CAMERA,
        ...
    },
    "shots": {
        SHOT_ID: SHOT,
        ...
    },
    "points": {
        POINT_ID: POINT,
        ...
    }
}

CAMERA: {
    "projection_type": "perspective",  # Can be perspective, brown, fisheye or equirectangular
    "width": NUMBER,                   # Image width in pixels
    "height": NUMBER,                  # Image height in pixels

    # Depending on the projection type more parameters are stored.
    # These are the parameters of the perspective camera.
    "focal": NUMBER,                   # Estimated focal length
    "k1": NUMBER,                      # Estimated distortion coefficient
    "k2": NUMBER,                      # Estimated distortion coefficient
}

SHOT: {
    "camera": CAMERA_ID,
    "rotation": [X, Y, Z],      # Estimated rotation as an angle-axis vector
    "translation": [X, Y, Z],   # Estimated translation
    "gps_position": [X, Y, Z],  # GPS coordinates in the reconstruction reference frame
    "gps_dop": METERS,          # GPS accuracy in meters
    "orientation": NUMBER,      # EXIF orientation tag (can be 1, 3, 6 or 8)
    "capture_time": SECONDS     # Capture time as a UNIX timestamp
}

POINT: {
    "coordinates": [X, Y, Z],      # Estimated position of the point
    "color": [R, G, B],            # Color of the point
}
"""

"""
reconstruction.meshed.json
[{
    'cameras': {'v2 unknown unknown 1920 1080 perspective 0':
                {'projection_type': 'perspective', 'width': 1920, 'height': 1080, 'focal': 0.8647151305270488,
                'k1': 0.04060214391621549, 'k2': -0.04060273810852096}},
    'shots': {
        'frame00016.png' : {
        'rotation': [1.5061234719524716, 0.06688721174244067, -0.030847050348337724],
         'translation': [-2.1535823328020456, 0.28345212194377944, 1.2491740134158436],
         'camera': 'v2 unknown unknown 1920 1080 perspective 0',
         'orientation': 1,
         'capture_time': 0.0,
         'gps_dop', 'gps_position',
         'vertices',     # only in meshed
         'faces' [[x, y, z, ...]  # only in meshed
         },
         ...
    },
     'points': {
        '1': {
            'color': [74.0, 43.0, 31.0],
            'coordinates': [-4.204454761953588, 11.796709404713068, 4.7276044200915]
        },
        ...
     },
     'reference_lla': {
        'latitude': 0.0,
        'longitude': 0.0,
        'altitude': 0
     }
}]
"""


def export_opensfm_camera(
        kapture_camera: kapture.Camera
) -> Dict[str, Any]:
    """
    Converts kapture camera to OpenSfM.
    OpenSfm propose 3 models of camera: perspective, fisheye, spherical.
    Perspective Camera of OpenSfM
    (see https://www.opensfm.org/docs/geometry.html#camera-models)

    OpenSfM camera axis:
        - The z-axis points forward
        - The y-axis points down
        - The x-axis points to the right

    OpenSfM Pose:
        OpenSfM, however, chooses not to store the “camera origin” in Pose objects.
        Instead, it stores the camera coordinates of the world origin in the translation field.
        Meaning Camera from World transformation.
        - rotation is represented as axis-angle vector.

    :param kapture_camera: camera kapture definition
    :return: camera definition as dictionary to save (as json, db record, ...)
    """
    corresponding_opensfm_perspective = [
        kapture.CameraType.SIMPLE_PINHOLE,
        kapture.CameraType.RADIAL,
        kapture.CameraType.SIMPLE_RADIAL
    ]

    if kapture_camera.camera_type not in corresponding_opensfm_perspective:
        raise ValueError(f'Unable to export camera of type "{kapture_camera.camera_type}" to OpenSfM')

    # SIMPLE_PINHOLE = w, h, fx, fy, cx, cy
    # SIMPLE_RADIAL  = w, h, f, cx, cy, k1
    # RADIAL         = w, h, f, cx, cy, k1, k2

    # note that principal point is ignored by openSfM
    largest_side_in_pixels = max(kapture_camera.camera_params[0], kapture_camera.camera_params[1])
    opensfm_camera = {
        'projection_type': 'perspective',
        'width': int(kapture_camera.camera_params[0]),
        'height': int(kapture_camera.camera_params[1]),
        # The focal length provided by the EXIF metadata divided by the sensor width
        'focal': kapture_camera.camera_params[2] / largest_side_in_pixels,
        'k1': 0.0,
        'k2': 0.0,
    }

    # Distortion model of OpenSfM match CameraType.RADIAL
    if kapture_camera.camera_type in [kapture.CameraType.SIMPLE_RADIAL, kapture.CameraType.RADIAL]:
        opensfm_camera['k1'] = kapture_camera.camera_params[5]

    if kapture_camera.camera_type == kapture.CameraType.RADIAL:
        opensfm_camera['k2'] = kapture_camera.camera_params[6]

    return opensfm_camera


def _export_opensfm_features_and_matches(image_filenames,
                                         keypoints_type,
                                         descriptors_type,
                                         kapture_data,
                                         kapture_root_dir,
                                         tar_handlers,
                                         opensfm_root_dir,
                                         disable_tqdm):
    """
    export features files (keypoints + descriptors) and matches
    """
    opensfm_features_suffix = '.features.npz'
    opensfm_features_dir_path = path.join(opensfm_root_dir, 'features')
    logger.info(f'exporting keypoint and descriptors to {opensfm_features_dir_path}')
    os.makedirs(opensfm_features_dir_path, exist_ok=True)

    if keypoints_type is None:
        keypoints_type = try_get_only_key_from_collection(kapture_data.keypoints)
    # if it's still None try again from matches
    if keypoints_type is None:
        keypoints_type = try_get_only_key_from_collection(kapture_data.matches)
    if descriptors_type is None:
        descriptors_type = try_get_only_key_from_collection(kapture_data.descriptors)

    for image_filename in tqdm(image_filenames, disable=disable_tqdm):
        opensfm_features = {}
        # look and load for keypoints in kapture
        if kapture_data.keypoints is not None and \
                keypoints_type is not None and \
                keypoints_type in kapture_data.keypoints and \
                image_filename in kapture_data.keypoints[keypoints_type]:
            kapture_keypoints_filepath = get_keypoints_fullpath(keypoints_type=keypoints_type,
                                                                kapture_dirpath=kapture_root_dir,
                                                                image_filename=image_filename,
                                                                tar_handler=tar_handlers)
            logger.debug(f'loading {kapture_keypoints_filepath}')
            kapture_keypoint = image_keypoints_from_file(kapture_keypoints_filepath,
                                                         dtype=kapture_data.keypoints[keypoints_type].dtype,
                                                         dsize=kapture_data.keypoints[keypoints_type].dsize)
            opensfm_features['points'] = kapture_keypoint

        # look and load for descriptors in kapture
        if kapture_data.descriptors is not None and \
                descriptors_type is not None and \
                descriptors_type in kapture_data.descriptors and \
                image_filename in kapture_data.descriptors[descriptors_type]:
            kapture_descriptor_filepath = get_descriptors_fullpath(descriptors_type=descriptors_type,
                                                                   kapture_dirpath=kapture_root_dir,
                                                                   image_filename=image_filename,
                                                                   tar_handler=tar_handlers)
            logger.debug(f'loading {kapture_descriptor_filepath}')
            kapture_descriptor = image_descriptors_from_file(kapture_descriptor_filepath,
                                                             dtype=kapture_data.descriptors[descriptors_type].dtype,
                                                             dsize=kapture_data.descriptors[descriptors_type].dsize)
            opensfm_features['descriptors'] = kapture_descriptor

        # writing opensfm feature file
        if len(opensfm_features) > 0:
            opensfm_features_filepath = path.join(opensfm_features_dir_path, image_filename + opensfm_features_suffix)
            logger.debug(f'writing {opensfm_features_filepath}')
            os.makedirs(path.dirname(opensfm_features_filepath), exist_ok=True)
            np.save(opensfm_features_filepath, opensfm_features)

    # export matches files
    if kapture_data.matches is not None and \
            keypoints_type is not None and \
            keypoints_type in kapture_data.matches:
        opensfm_matches_suffix = '_matches.pkl.gz'
        opensfm_matches_dir_path = path.join(opensfm_root_dir, 'matches')
        os.makedirs(opensfm_matches_dir_path, exist_ok=True)
        logger.info(f'exporting matches to {opensfm_matches_dir_path}')
        opensfm_pairs = {}
        for image_filename1, image_filename2 in kapture_data.matches[keypoints_type]:
            opensfm_pairs.setdefault(image_filename1, []).append(image_filename2)

        for image_filename1 in tqdm(image_filenames, disable=disable_tqdm):
            opensfm_matches = {}
            opensfm_matches_filepath = path.join(opensfm_matches_dir_path, image_filename1 + opensfm_matches_suffix)
            logger.debug(f'loading matches for {image_filename1}')
            for image_filename2 in opensfm_pairs.get(image_filename1, []):
                # print(image_filename1, image_filename2)
                kapture_matches_filepath = get_matches_fullpath(
                    (image_filename1, image_filename2),
                    keypoints_type=keypoints_type,
                    kapture_dirpath=kapture_root_dir,
                    tar_handler=tar_handlers)
                kapture_matches = image_matches_from_file(kapture_matches_filepath)
                opensfm_matches[image_filename2] = kapture_matches[:, 0:2].astype(np.int)

            os.makedirs(path.dirname(opensfm_matches_filepath), exist_ok=True)
            with gzip.open(opensfm_matches_filepath, 'wb') as f:
                pickle.dump(opensfm_matches, f)


def export_opensfm(
        kapture_root_dir: str,
        opensfm_root_dir: str,
        force_overwrite_existing: bool = False,
        images_export_method: TransferAction = TransferAction.copy,
        keypoints_type: Optional[str] = None,
        descriptors_type: Optional[str] = None,
) -> None:
    """
    Export the kapture data to an openSfM format

    :param kapture_root_dir: full path to the top kapture directory
    :param opensfm_root_dir: path of the directory where to store the data in openSfM format
    :param force_overwrite_existing: if true, will remove existing openSfM data without prompting the user.
    :param images_export_method:
    """

    disable_tqdm = logger.getEffectiveLevel() > logging.INFO  # don't display tqdm for non-verbose levels
    # load reconstruction
    with get_all_tar_handlers(kapture_root_dir) as tar_handlers:
        kapture_data = kapture_from_dir(kapture_root_dir, tar_handlers=tar_handlers)

        # export cameras
        opensfm_cameras = {}
        for cam_id, kapture_camera in kapture_data.cameras.items():
            opensfm_cameras[cam_id] = export_opensfm_camera(kapture_camera)

        # export shots
        opensfm_shots = {}
        for timestamp, camera_id, image_filename in tqdm(kapture.flatten(kapture_data.records_camera),
                                                         disable=disable_tqdm):
            # retrieve pose (if there is one).
            # opensfm_shots = {image_filename: shot}
            # shot = {camera , rotation, translation, capture_time, gps_position, ...}
            opensfm_shot = {
                'capture_time': 0,  # in ms != timestamp
                'camera': camera_id,
            }
            if (timestamp, camera_id) in kapture_data.trajectories.key_pairs():
                pose = kapture_data.trajectories[timestamp, camera_id]
                rotation_vector = quaternion.as_rotation_vector(pose.r)
                translation_vector = pose.t.flatten()
                opensfm_shot.update({
                    'rotation': rotation_vector.tolist(),
                    'translation': translation_vector.tolist()
                })
            opensfm_shots[image_filename] = opensfm_shot

        # pack it
        opensfm_reconstruction = {
            'cameras': opensfm_cameras,
            'shots': opensfm_shots,
        }

        # images
        logger.info(f'writing image files "{path.join(opensfm_root_dir, "images")}".')
        image_filenames = [f for _, _, f in kapture.flatten(kapture_data.records_camera)]
        kapture_image_file_paths = [get_record_fullpath(kapture_root_dir, image_filename)
                                    for image_filename in image_filenames]
        opensfm_image_file_paths = [path.join(opensfm_root_dir, 'images', image_filename)
                                    for image_filename in image_filenames]
        transfer_files_from_dir(
            source_filepath_list=kapture_image_file_paths,
            destination_filepath_list=opensfm_image_file_paths,
            force_overwrite=force_overwrite_existing,
            copy_strategy=images_export_method,
        )

        _export_opensfm_features_and_matches(image_filenames, keypoints_type, descriptors_type,
                                             kapture_data, kapture_root_dir, tar_handlers,
                                             opensfm_root_dir, disable_tqdm)

        # export 3D-points files
        if kapture_data.points3d is not None:
            logger.info('exporting points 3-D')
            opensfm_reconstruction['points'] = {}
            for i, (x, y, z, r, g, b) in tqdm(enumerate(kapture_data.points3d), disable=disable_tqdm):
                opensfm_reconstruction['points'][i] = {
                    'coordinates': [x, y, z],
                    'color': [r, g, b]
                }

    # write json files #################################################################################################
    os.makedirs(opensfm_root_dir, exist_ok=True)
    # write reconstruction.json
    opensfm_reconstruction_filepath = path.join(opensfm_root_dir, 'reconstruction.json')
    logger.info(f'writing reconstruction file "{opensfm_reconstruction_filepath}".')
    with open(opensfm_reconstruction_filepath, 'wt') as f:
        json.dump([opensfm_reconstruction], f, indent=4)

    # write camera_models.json
    opensfm_cameras_filepath = path.join(opensfm_root_dir, 'camera_models.json')
    logger.info(f'writing camera models file "{opensfm_cameras_filepath}".')
    with open(opensfm_cameras_filepath, 'wt') as f:
        json.dump(opensfm_cameras, f, indent=4)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Logging specific to kapture
"""

import argparse
import logging

logging.basicConfig(format='%(levelname)-8s::%(name)s: %(message)s')


class VerbosityParser(argparse.Action):
    """ accept debug, info, ... or theirs corresponding integer value formatted as string."""

    def __call__(self, parser, namespace, values, option_string=None):
        assert isinstance(values, (int, str))
        try:  # in case it represent an int, directly get it
            values = int(values)
        except ValueError:  # else ask logging to sort it out
            values = logging.getLevelName(values.upper())
        setattr(namespace, self.dest, values)


def getLogger():
    """
    Get the default kapture logger.

    :return: logger
    """
    return logging.getLogger('kapture')



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports an openmvg file in the kapture format.
"""

import argparse
import logging
import sys
import os.path as path

# kapture
import path_to_kapture  # enables import kapture  # noqa: F401
import kapture.utils.logging
from kapture.io.records import TransferAction
# openmvg
from kapture.converter.openmvg.import_openmvg import import_openmvg
from kapture.converter.openmvg.openmvg_commons import OPENMVG_DEFAULT_JSON_FILE_NAME

logger = logging.getLogger('openmvg')


def import_openmvg_command_line() -> None:
    """
    Do the openmvg to kapture import using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='Imports from openMVG JSON file to Kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-y', '--force', action='store_true', default=False,
                        help='silently delete kapture data if already exists.')
    # create the parser for the import command #########################################################################
    parser.add_argument('-i', '--openmvg',
                        help='path to openMVG directory, will automatically look for sfm_data, regions and pair files.')
    parser.add_argument('-s', '--sfm_data',
                        help='path to openMVG sfm_data data file.')
    parser.add_argument('-r', '--regions',
                        help='path to openMVG directory containing region files (feat, desc).')
    parser.add_argument('-m', '--matches',
                        help='path to openMVG matches file (eg. matches.f.txt')
    parser.add_argument('-o', '-k', '--kapture', required=True,
                        help='top directory where to save Kapture files.')
    parser.add_argument('--image_action', default='root_link', type=TransferAction,
                        help=f'''what to do with images:
        {TransferAction.root_link.name}: link to the root of the images directory (default) ;
        {TransferAction.link_absolute.name}: absolute individual file link ;
        {TransferAction.link_relative.name}: relative individual file link ;
        {TransferAction.copy.name}: copy file instead of creating link ;
        {TransferAction.move.name}: move file instead of creating link ;
        {TransferAction.skip.name}: do not create links
                                      ''')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    # argument logic.
    if args.openmvg:
        args.openmvg = path.abspath(args.openmvg)
        # automatically look for sfm_data, regions and pair files if not specified
        if args.sfm_data is None:
            args.sfm_data = path.join(args.openmvg, OPENMVG_DEFAULT_JSON_FILE_NAME)
        if args.regions is None:
            args.regions = args.openmvg
        if args.matches is None:
            args.matches = path.join(args.openmvg, 'matches.f.txt')

    # normalizes paths
    if args.sfm_data is not None:
        args.sfm_data = path.normpath(path.abspath(args.sfm_data))
    if args.regions is not None:
        args.regions = path.normpath(path.abspath(args.regions))
    if args.matches is not None:
        args.matches = path.normpath(path.abspath(args.matches))
        assert args.matches.endswith('.txt')

    # sanity check
    if all(i is None for i in [args.sfm_data, args.regions, args.matches]):
        raise ValueError('Need openMVG files as input.')

    import_openmvg(args.sfm_data,
                   args.regions,
                   args.matches,
                   args.kapture,
                   args.image_action,
                   args.force)


if __name__ == '__main__':
    try:
        import_openmvg_command_line()
    except (AssertionError, LookupError, OSError):
        logger.error('Fatal error', exc_info=True)



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Import a rosbag with images from several camera and position associated to the images recorded.
Works by default with the RealSense T265 camera.
Must have ROS installed, as well as rosbag module.
"""

from collections import OrderedDict
import logging
import os
import os.path as path
import sys
from typing import Any, Dict, List, Optional, Union
import numpy as np
import PIL.Image as PILImage
import quaternion
from tqdm import tqdm
# ros
import geometry_msgs.msg
import rosbag
import rospy.rostime
from rospy.rostime import Duration
# kapture
import kapture
from kapture.core.flatten import flatten
from kapture.core.Rigs import Rigs
from kapture.core.Sensors import Sensors
import kapture.io.csv as kcsv
import kapture.io.structure
from kapture.io.records import get_image_fullpath
from kapture.utils.Collections import LimitedDictionary
from kapture.utils.paths import path_secure


rotation_cam_kapture_from_cam_ros = np.array([
    [0, -1, 0],
    [0, 0, -1],
    [1, 0, 0]
])

pose_kapture_from_ros = kapture.PoseTransform(r=quaternion.from_rotation_matrix(rotation_cam_kapture_from_cam_ros))
pose_ros_from_kapture = pose_kapture_from_ros.inverse()


class ImageInfo:
    """
    Image info found in the Rosbag
    """

    def __init__(self,
                 filename: str,
                 timestamp: rospy.rostime.Time,
                 camera_name: str):
        self.filename = filename
        self.timestamp = timestamp
        self.camera_name = camera_name


class PositionInfo:
    """
    Position info found in the Rosbag
    """

    def __init__(self,
                 timestamp: rospy.rostime.Time,
                 pose6d: geometry_msgs.msg.Pose):
        self.timestamp = timestamp
        self.pose6d = pose6d


def _extract_img_buf(msg) -> np.ndarray:
    """
    Extracts the image as numpy one dimension array from the ROS message
    """
    # TODO: Better image decoding:
    #  cv_bridge (depend on OpenCV)
    #  ros_numpy: no dependency on OpenCV but not natively installed on melodic
    # The below code is adapted from its Image.py file and will work for types like:
    # rgb8, rgba8, rgb16, rgba16, bgr8, bgra8, bgr16, bgra16, mono8, mono16
    # and the bayer formats but not the OpenCV CvMat types.
    dtype = np.uint8
    if len(msg.encoding) > 0:
        if msg.encoding.endswith("8"):
            dtype = np.uint8
        elif msg.encoding.endswith("16"):
            dtype = np.uint16
        else:
            raise ValueError(f'Unsupported image encoding {msg.encoding}')
    nb_channels = 1
    if msg.step > 0:
        nb_channels = int(msg.step / msg.width)
    if nb_channels > 1:
        img_buf = np.frombuffer(msg.data, dtype=dtype).reshape(msg.height, msg.width, nb_channels)
    else:
        img_buf = np.frombuffer(msg.data, dtype=dtype).reshape(msg.height, msg.width)
    return img_buf


def _check_timestamp_delta(images_stamp: Dict[Any, rospy.rostime.Time]) -> None:
    # The images are taken every 0.033 second
    # Check all images have the same time stamp modulo epsilon
    for topic1, stamp1 in images_stamp.items():
        for topic2, stamp2 in images_stamp.items():
            delta = (stamp1 - stamp2).to_sec()
            assert (abs(delta) < 0.001), "{topic1} and {topic2} images must have the same timestamp"


class RosBagImporter:
    """
    A importer of ROS bags with multi cam images data and odometry.
    The join between the two types of data is made on their data timestamps.
    """

    def __init__(self,
                 rosbag_path: str,
                 rigs: Optional[Rigs],
                 sensors: Sensors,
                 kapture_path: str,
                 force_overwrite_existing: bool = False) -> None:
        """

        :param rosbag_path: full path to the rosbag file
        :param rigs: rigs of the sensors
        :param sensors: sensors definition used for the capture
        :param kapture_path: full path to the top kapture directory to save
        :param force_overwrite_existing: silently overwrite kapture files if already exists
        """
        if not path.isfile(rosbag_path):
            raise ValueError(f'Rosbag file {rosbag_path} does not exist')
        self._rosbag_path = rosbag_path
        self._rigs = rigs
        self._sensors = sensors
        self._kapture_path = kapture_path
        self.logger = logging.getLogger('rosbag')
        self.logger.info(f'Reading rosbag file {rosbag_path} and exporting as Kapture format into {kapture_path}')
        os.makedirs(kapture_path, exist_ok=True)
        kapture.io.structure.delete_existing_kapture_files(kapture_path, force_overwrite_existing)
        self._images_full_path = get_image_fullpath(kapture_path)
        self._image_directory_path = OrderedDict()  # Should be OrderedDict[str, str]
        self._image_topic_to_cam_id = dict()
        # Keys = timestamp, values odometer poses
        self._last_poses = LimitedDictionary(20)
        self.images_info = list()  # Of ImageInfo: type annotation is not supported in 3.6
        self.poses_info = list()  # Of PositionInfo
        self._image_count_per_camera = -1
        self._saved_number = 0

    def _check_bag_topics(self, bag: rosbag.bag.Bag, odometry_topic: Optional[str], image_topics: List[str]):
        bag_info = bag.get_type_and_topic_info()
        all_topics = list(bag_info[1].keys())
        self.logger.debug(f'Topics found {all_topics}')
        # Check topics
        if odometry_topic and odometry_topic not in all_topics:
            raise ValueError(f'Missing topic {odometry_topic} in Rosbag')
        for image_topic in image_topics:
            if image_topic not in all_topics:
                raise ValueError(f'Missing image topic {image_topic} in Rosbag')
        message_count = bag.get_message_count()
        self.logger.info(f"{message_count} messages in total")
        self._image_count_per_camera = -1
        for topic in all_topics:
            stats = bag_info[1][topic]
            count = stats[1]
            self.logger.info(f'In topic {topic:30s} {count:#10d} messages of type {stats[0]}')
            # Make sure we have the same number of images for all camera
            if topic in image_topics:
                if self._image_count_per_camera >= 0:
                    assert count == self._image_count_per_camera, "All cameras have the same number of images"
                self._image_count_per_camera = count

    def _create_images_directories(self, camera_identifiers, image_topics):
        self._image_directory_path.clear()
        self._image_topic_to_cam_id.clear()
        for (image_topic, camera_id) in zip(image_topics, camera_identifiers):
            self._image_topic_to_cam_id[image_topic] = camera_id
            self.logger.info(f'images from {image_topic} mapped to camera {camera_id}')
            dir_path = path.join(self._images_full_path, camera_id)
            self._image_directory_path[image_topic] = dir_path
            os.makedirs(dir_path, exist_ok=True)
        self.logger.info(f'Saving images into {" ".join(dir_path for dir_path in self._image_directory_path.values())}')

    def _save_image(self,
                    image_bitmap: np.ndarray,
                    image_directory_path: str,
                    image_number: int,
                    timestamp: rospy.rostime.Time) -> str:
        """
        Save the image.

        :param image_bitmap: the image bytes
        :param image_directory_path: directory where to save the image, under the top kapture directory
        :param image_number: image sequence number
        :param timestamp: image time stamp to set to the file
        :return: relative path (to the kapture path) of the file
        """
        # Define image file path
        image_path = os.path.join(image_directory_path, f'frame_{image_number:09d}.jpg')
        img = PILImage.fromarray(image_bitmap)
        img.save(image_path)
        self._saved_number += 1
        # Set the capture time to the file saved
        taken_time = timestamp.to_sec()
        os.utime(image_path, (taken_time, taken_time))
        return path.relpath(image_path, self._images_full_path)

    def _find_pose(self, image_stamp: rospy.rostime.Time) -> Optional[geometry_msgs.msg.Pose]:
        """
        Find the pose that has the smallest time difference with the image timestamp

        :param image_stamp: an image time stamp
        :return: a geometric pose if found, none otherwise
        """
        if len(self._last_poses) == 0:
            return None
        chosen_stamp = None
        smallest_delta = Duration(sys.maxsize)  # Very big delta
        for stamp in self._last_poses.keys():
            delta = abs(image_stamp-stamp)
            if delta < smallest_delta:
                smallest_delta = delta
                chosen_stamp = stamp
        return self._last_poses[chosen_stamp]

    def import_multi_camera(self, odometry_topic: Optional[str],  # noqa: C901
                            image_topics: Union[str, List[str]],
                            camera_identifiers: Union[str, List[str]],
                            save_all_positions: bool = True,
                            find_image_position: bool = True,
                            percent: int = 100):
        """
        Import the rosbag data. Save the images on disk.
        The image topics list and camera identifiers list must be matching lists.

        :param odometry_topic: the odometry topic to use to compute the trajectory.
        :param image_topics: image topic(s) to import
        :param camera_identifiers: camera identifier(s) corresponding to the image topic(s)
        :param save_all_positions: save all positions from the odometry topic in the trajectory
        :param find_image_position: find the closest position for the image and add it in the trajectory
        :param percent: percentage of images to keep.
        """

        # The image messages have the following structure:
        #
        # std_msgs/Header header
        # uint32 height
        # uint32 width
        # string encoding
        # uint8 is_bigendian
        # uint32 step
        # uint8[] data
        #

        # Check that we have image topics
        if isinstance(image_topics, str):
            image_topics = [image_topics]
        if image_topics is None or len(image_topics) == 0:
            self.logger.fatal('Please provide image topics')
            raise ValueError('Missing image topic')
        # Check that we have as many camera identifiers as image topics
        if isinstance(camera_identifiers, str):
            camera_identifiers = [camera_identifiers]
        nb_image_topics = len(image_topics)
        if nb_image_topics != len(camera_identifiers):
            self.logger.fatal(f'Please provide an equal number of image topics and camera identifiers:'
                              f' {len(camera_identifiers)} cameras for {nb_image_topics} image topics.')
            raise ValueError('Unequal number of image topics and camera identifiers')
        topics_to_import = image_topics.copy()
        # Check that we have an odometry topic
        if odometry_topic and len(odometry_topic) == 0:
            odometry_topic = None
        if odometry_topic is None:
            self.logger.info('No odometry topic')
            if find_image_position:
                self.logger.info('Will not find position for the images without odometry')
                find_image_position = False
        else:
            topics_to_import.append(odometry_topic)

        # Make sure the camera identifiers are defined in the list of sensors
        for camera_id in camera_identifiers:
            if self._sensors.get(camera_id) is None:
                raise ValueError(f'Camera identifier {camera_id} is not defined in {list(self._sensors.keys())}')
        # Create images directories
        self._create_images_directories(camera_identifiers, image_topics)

        with rosbag.Bag(self._rosbag_path) as bag:
            self._check_bag_topics(bag, odometry_topic, image_topics)

            # Read the messages
            num_msgs = 1
            image_number = 1
            self._saved_number = 0
            images_stamp: Dict[Any, rospy.rostime.Time] = {}  # images time stamp of type rospy.rostime.Time per topic
            images_buffer: Dict[str, np.ndarray] = {}  # the images read as byte array per topic : Dict[str, np.ndarray]
            dir_last_image_time = {}  # image directory -> time of the last image written : Dict[str, float]
            count_to_skip = int(100 / percent) if percent < 100 else 1
            self.logger.info(f'Reading {self._image_count_per_camera} images per camera')
            if percent < 100:
                self.logger.info(f' Keeping {percent}% of the images')
            if not self.logger.isEnabledFor(logging.DEBUG):
                progress_bar = tqdm(total=self._image_count_per_camera)
            else:
                progress_bar = None
            for topic, msg, t in bag.read_messages(topics=topics_to_import):
                # Consider the data timestamp, not the recording timestamp t
                stamp = msg.header.stamp
                if topic == odometry_topic:
                    self.logger.debug(f"{num_msgs:5d} STAMP='{stamp.secs}.{stamp.nsecs}'"
                                      f" ODOM msg.poseX='{msg.pose.pose.position.x}'")
                    self._last_poses[stamp] = msg.pose.pose
                    if save_all_positions:
                        self.poses_info.append(PositionInfo(stamp, msg.pose.pose))
                elif topic in image_topics:
                    self.logger.debug(f"{num_msgs:5d} STAMP='{stamp.secs}.{stamp.nsecs}'"
                                      f" {topic} HxW={msg.height}x{msg.width}")
                    images_stamp[topic] = stamp
                    images_buffer[topic] = _extract_img_buf(msg)
                else:
                    self.logger.debug(f"{num_msgs:5d} Ignoring message of topic '{topic}'")
                # If we have all images, save them on disk
                if len(images_stamp) == nb_image_topics:
                    _check_timestamp_delta(images_stamp)
                    pose6d = self._find_pose(stamp) if find_image_position else None
                    if pose6d or not find_image_position:
                        # Save only some images
                        if ((image_number-1) % count_to_skip) == 0:
                            for image_topic in image_topics:
                                img = images_buffer[image_topic]
                                image_directory = self._image_directory_path[image_topic]
                                stamp = images_stamp[image_topic]
                                try:
                                    img_name = path_secure(self._save_image(img, image_directory, image_number, stamp))
                                except SystemError:
                                    self.logger.exception(f"Could not save image number {image_number}"
                                                          f" from topic {image_topic}")
                                else:
                                    dir_last_image_time[image_directory] = stamp.to_sec()
                                    self.images_info.append(
                                        ImageInfo(img_name, stamp, self._image_topic_to_cam_id[image_topic])
                                    )
                            if pose6d:
                                self.poses_info.append(PositionInfo(stamp, pose6d))
                        else:
                            self.logger.debug(f'    Skipping image number {image_number}')
                        image_number += 1
                        images_stamp.clear()
                        images_buffer.clear()
                        progress_bar and progress_bar.update(1)
                    # Else wait until we have the first pose
                num_msgs += 1
            progress_bar and progress_bar.close()
            for image_dir, dir_time in dir_last_image_time.items():
                os.utime(image_dir, (dir_time, dir_time))
            self.logger.info(f'Saved {self._saved_number} images')

    def save_to_kapture(self, trajectory_rig_id: Optional[str] = None) -> None:
        """
        Save the data in kapture format.

        :param trajectory_rig_id: the rig identifier of the trajectory points
        """
        # Convert pose info to trajectories
        if len(self.poses_info) > 0 and trajectory_rig_id is None:
            raise ValueError("Must provide rig identifier for trajectory")
        trajectories = kapture.Trajectories() if len(self.poses_info) > 0 else None
        for pose_info in self.poses_info:
            t = pose_info.timestamp.to_nsec()
            ros_translation = pose_info.pose6d.position
            translation = [ros_translation.x,
                           ros_translation.y,
                           ros_translation.z]
            ros_rotation = pose_info.pose6d.orientation
            rotation = np.quaternion(ros_rotation.w,
                                     ros_rotation.x,
                                     ros_rotation.y,
                                     ros_rotation.z)
            # Transform the pose from the ROS body coordinate system defined here
            # https://www.ros.org/reps/rep-0103.html#axis-orientation
            # to the Kapture coordinate system

            # ros pose seems to be the inverse of the extrinsic matrix
            # i.e world position and rig orientation with respect to the world axis
            pose6d = kapture.PoseTransform.compose([pose_kapture_from_ros,
                                                    kapture.PoseTransform(rotation, translation).inverse(),
                                                    pose_ros_from_kapture])
            trajectories[(t, trajectory_rig_id)] = pose6d
        self.logger.info(f'Saving {len(list(flatten(trajectories)))} poses')
        # Convert image info to kapture image
        records_camera = kapture.RecordsCamera()
        for image_info in self.images_info:
            t = image_info.timestamp.to_nsec()
            records_camera[(t, image_info.camera_name)] = image_info.filename
        self.logger.info(f'Saving {len(list(flatten(records_camera)))} camera records')

        kapture_data = kapture.Kapture(rigs=self._rigs, sensors=self._sensors,
                                       records_camera=records_camera, trajectories=trajectories)
        self.logger.info(f'Saving to kapture {self._kapture_path}')
        kcsv.kapture_to_dir(self._kapture_path, kapture_data)
        self.logger.info('Done')



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Import all types of UTBM sensors information files to make the sensors list.
These files are published on github at:
https://github.com/epan-utbm/utbm_robocar_dataset/tree/baselines/camera_info

The files have been produced by the ROS camera_calibration package:
https://wiki.ros.org/camera_calibration
"""

from os import path
from typing import List, Union
import yaml

from kapture.core.Sensors import CameraType, Camera, Sensors

#  UTBM specific constants values

# Bumblebee2 cameras ROS topics and identifiers in same order:
TOPICS_BB2 = ['/BB2_9211442/left/image_raw',
              '/BB2_9211442/right/image_raw']
BB2_CAMERA_IDENTIFIERS = ['bb2_left', 'bb2_right']

# Bumblebee XB3 cameras ROS topics and identifiers in same order:
TOPICS_XB3 = ['/BBX3_7140017/camera/center/image_raw',
              '/BBX3_7140017/camera/left/image_raw',
              '/BBX3_7140017/camera/right/image_raw']
XB3_CAMERA_IDENTIFIERS = ['bb_xb3_center', 'bb_xb3_left', 'bb_xb3_right']


def import_ros_camera_calibration(camera_info_file_path: str) -> Camera:
    """
    Parse a ROS camera calibration file to insert in the sensor list.

    :param camera_info_file_path: path to the camera info file
    :return: the camera name and its parameters
    """
    camera_info_file_path = path.abspath(camera_info_file_path)
    with open(camera_info_file_path) as file:
        camera_info = yaml.full_load(file)
        camera_name = camera_info['camera_name']
        w, h = camera_info['image_width'], camera_info['image_height']
        camera_matrix = camera_info['camera_matrix']['data']
        fx, cx, fy, cy = camera_matrix[0], camera_matrix[2], camera_matrix[4], camera_matrix[5]
        # Will use opencv model with parameters: w, h, fx, fy, cx, cy, k1, k2, p1, p2
        distortion_coefficients = camera_info['distortion_coefficients']['data']
        # Make sure k3 is null
        assert(distortion_coefficients[4] == float(0.0))
        camera_parameters = [w, h, fx, fy, cx, cy] + distortion_coefficients[0:4]
        camera = Camera(CameraType.OPENCV, camera_parameters, camera_name)
        return camera


def import_utbm_sensors(camera_info_file_paths: Union[str, List[str]]) -> Sensors:
    """
    Import all camera info.

    :param camera_info_file_paths: the list of camera info file paths
    :return: kapture Sensors with all camera defined
    """
    sensors = Sensors()
    if isinstance(camera_info_file_paths, str):
        camera_info_file_paths = [camera_info_file_paths]
    for camera_info_file in camera_info_file_paths:
        sensor = import_ros_camera_calibration(camera_info_file)
        sensors[sensor.name] = sensor
    return sensors



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

import numpy as np


class Points3d(np.ndarray):
    """
    3D points. This can be used like this:

    - 3-D points is a Nx6 a numpy array : 1 point per row x 6 cols (x,y,z,r,g,b)
    - color (rgb) is coded from 0 to 255
    """
    XYZRGB = 6

    def __new__(cls, input_array=None):
        """
        Creates the ndarray instance of our type, given the usual ndarray input arguments.
        This will call the standard ndarray constructor, but return an object of our type.

        :param input_array
        """
        if input_array is None:
            # default constructor, empty shape
            empty_shape = (0, Points3d.XYZRGB)
            obj = super(Points3d, cls).__new__(cls, empty_shape).view(cls)
        elif isinstance(input_array, np.ndarray):
            # Input array is an already formed ndarray instance
            # We first cast to be our class type
            obj = np.asarray(input_array, dtype=np.float64).view(cls)
        elif isinstance(input_array, list):
            # convert to numpy array and cast to our class type
            obj = np.array(input_array, dtype=np.float64).view(cls)
        else:
            raise TypeError('Unknown type to convert 3-D points from')

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        if obj is None or (self.shape and self.shape == (6,)):  # Second part is for the IDE
            return
        # from casting or view, check the shape is correct
        if not self.shape or len(self.shape) < 2 or not self.shape[1] == Points3d.XYZRGB:
            raise ValueError(
                '3D points are expected to be Nx6 shape ({})'.format(self.shape))

    def __getitem__(self, items):
        # with __array_finalize__ strict shape checking, getting a slice of Points3d can be an issue e.g p[:, 0:3] fails
        # this function tries to slice the Points3d normally.
        # If __array_finalize__ fails: e.g shape[1] is not 6 then it will return it as a slice of a normal numpy array
        try:
            return super(Points3d, self).__getitem__(items)
        except ValueError:
            return self.as_array().__getitem__(items)

    def __bool__(self):
        """
        Checks for emptiness

        :return: True if not empty. """
        return self.shape[0] > 0

    def as_array(self):
        """ cast the Points3d to numpy array. """
        return self.view(dtype=np.float64, type=np.ndarray)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script imports an NVM model in the kapture format.

VisualSFM saves SfM workspaces into NVM files, which contain input image paths and multiple 3D models.
Below is the format description

NVM_V3 [optional calibration]                        # file version header
<Model1> <Model2> ...                                # multiple reconstructed models
<Empty Model containing the unregistered Images>     # number of camera > 0, but number of points = 0
<0>                                                  # 0 camera to indicate the end of model section
<Some comments describing the PLY section>
<Number of PLY files> <List of indices of models that have associated PLY>

The [optional calibration] exists only if you use "Set Fixed Calibration" Function
FixedK fx cx fy cy

Each reconstructed <model> contains the following
<Number of cameras> <List of cameras>
<Number of 3D points> <List of points>

The cameras and 3D points are saved in the following format
<Camera> = <File name> <focal length> <quaternion WXYZ> <camera center> <radial distortion> 0
<Point>  = <XYZ> <RGB> <number of measurements> <List of Measurements>
<Measurement> = <Image index> <Feature Index> <xy>
"""

import logging
import os
import os.path as path
import PIL
from PIL import Image
import numpy as np
import quaternion
import argparse
from typing import List, Optional, Set

# kapture
import path_to_kapture  # noqa: F401
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
from kapture.io.records import TransferAction, import_record_data_from_dir_auto
import kapture.io.features


logger = logging.getLogger('nvm')
MODEL = kapture.CameraType.SIMPLE_RADIAL
LOCAL_FEATURE_TYPE = 'sift'


def import_nvm(nvm_file_path: str,
               nvm_images_path: str,
               kapture_path: str,
               filter_list_path: Optional[str],
               ignore_trajectories: bool,
               add_reconstruction: bool,
               force_overwrite_existing: bool = False,
               images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Imports nvm data to kapture format.

    :param nvm_file_path: path to nvm file
    :param nvm_images_path: path to NVM images directory.
    :param kapture_path: path to kapture root directory.
    :param filter_list_path: path to the optional file containing a list of images to process
    :param ignore_trajectories: if True, will not create trajectories
    :param add_reconstruction: if True, will add observations, keypoints and 3D points.
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """

    # TODO implement [optional calibration]
    # doc : http://ccwu.me/vsfm/doc.html#nvm
    os.makedirs(kapture_path, exist_ok=True)
    delete_existing_kapture_files(kapture_path, force_erase=force_overwrite_existing)

    logger.info('loading all content...')
    # if there is a filter list, parse it
    # keep it as Set[str] to easily find images
    if filter_list_path:
        with open(filter_list_path) as file:
            file_content = file.readlines()
        # remove end line char and empty lines
        filter_list = {line.rstrip() for line in file_content if line != '\n'}
    else:
        filter_list = None

    # now do the nvm
    with open(nvm_file_path) as file:
        nvm_content = file.readlines()
    # remove end line char and empty lines
    nvm_content = [line.rstrip() for line in nvm_content if line != '\n']
    # only NVM_V3 is supported
    assert nvm_content[0] == "NVM_V3"
    # offset represents the line pointer
    offset = 1
    # camera_id_offset keeps tracks of used camera_id in case of multiple reconstructed models
    camera_id_offset = 0
    # point_id_offset keeps tracks of used point_id in case of multiple reconstructed models
    point_id_offset = 0

    cameras = kapture.Sensors()
    images = kapture.RecordsCamera()
    trajectories = kapture.Trajectories() if not ignore_trajectories else None
    if add_reconstruction:
        observations = kapture.Observations()
        keypoints = {LOCAL_FEATURE_TYPE: kapture.Keypoints(LOCAL_FEATURE_TYPE, np.float32, 2)}
        points3d = []
    else:
        observations = None
        keypoints = None
        points3d = None

    # break if number of cameras == 0 or reached end of file
    while True:
        # <Model1> <Model2> ...
        # Each reconstructed <model> contains the following
        # <Number of cameras> <List of cameras>
        # <Number of 3D points> <List of points>
        # In practice,
        # <Number of cameras>
        # <List of cameras>, one per line
        # <Number of 3D points>
        # <List of points>, one per line
        number_of_cameras = int(nvm_content[offset])
        offset += 1
        if number_of_cameras == 0:  # a line with <0> signify the end of models
            break

        logger.debug('importing model cameras...')
        # parse all cameras for current model
        image_idx_to_image_name = parse_cameras(number_of_cameras,
                                                nvm_content,
                                                offset,
                                                camera_id_offset,
                                                filter_list,
                                                nvm_images_path,
                                                cameras,
                                                images,
                                                trajectories)
        offset += number_of_cameras
        camera_id_offset += number_of_cameras

        # parse all points3d
        number_of_points = int(nvm_content[offset])
        offset += 1
        if points3d is not None and number_of_points > 0:
            assert keypoints is not None
            assert observations is not None
            logger.debug('importing model points...')
            parse_points3d(kapture_path,
                           number_of_points,
                           nvm_content,
                           offset,
                           point_id_offset,
                           image_idx_to_image_name,
                           filter_list,
                           points3d,
                           keypoints[LOCAL_FEATURE_TYPE],
                           observations)

        point_id_offset += number_of_points
        offset += number_of_points
        # reached end of file?
        if offset >= len(nvm_content):
            break

    # do not export values if none were found.
    if points3d is not None:
        points3d = kapture.Points3d(points3d)

    # import (copy) image files.
    logger.info('import image files ...')
    images_filenames = [f for _, _, f in kapture.flatten(images)]
    import_record_data_from_dir_auto(nvm_images_path, kapture_path, images_filenames, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(sensors=cameras, records_camera=images, trajectories=trajectories,
                                       points3d=points3d, keypoints=keypoints, observations=observations)
    logger.info('writing imported data...')
    kapture_to_dir(kapture_path, imported_kapture)


def parse_cameras(number_of_cameras: int,
                  nvm_content: List[str],
                  offset: int,
                  camera_id_offset: int,
                  filter_list: Optional[Set[str]],
                  nvm_images_path: str,
                  cameras: kapture.Sensors,
                  images: kapture.RecordsCamera,
                  trajectories: Optional[kapture.Trajectories]) -> List[str]:
    """
    Parse the <List of cameras> section
    Fill cameras, images, trajectories in place.
    Image files must exist to be able to retrieve height and width.

    :param number_of_cameras: number of cameras to process
    :param nvm_content: content of NVM file
    :param offset: number of characters to skip while reading every line
    :param camera_id_offset:
    :param filter_list: optional list of images to process
    :param nvm_images_path: path to NVM images directory
    :param cameras: kapture cameras to extend
    :param images: kapture images to extend
    :param trajectories: kapture trajectories to extend
    :return: list of images with position = index
    """
    image_idx_to_image_name = []
    # parse all cameras
    for i in range(0, number_of_cameras):
        line = nvm_content[i + offset].split()
        timestamp = i + camera_id_offset
        camera_id = f'sensor{timestamp}'
        image_file_name = line[0]
        image_idx_to_image_name.append(image_file_name)
        if filter_list is not None and image_file_name not in filter_list:
            # file_name is not in the list, do not add it
            continue

        focal_length = float(line[1])
        quaternion_wxyz = quaternion.from_float_array([float(v) for v in line[2:6]])
        camera_center = np.array([float(v) for v in line[6:9]])
        # https://github.com/colmap/colmap/blob/67e96894d4beed7cc93f1c0755a98d3664f85e63/src/base/reconstruction.cc#L891
        radial_distortion = -float(line[9])  # SIGN !

        try:
            # lazy open
            with Image.open(path.join(nvm_images_path, image_file_name)) as im:
                width, height = im.size
        except (OSError, PIL.UnidentifiedImageError):
            # It is not a valid image: skip it
            logger.info(f'Skipping invalid image file {image_file_name}')
            continue

        translation = - np.matmul(quaternion.as_rotation_matrix(quaternion_wxyz), camera_center)
        pose = kapture.PoseTransform(quaternion_wxyz, translation)

        camera = kapture.Camera(MODEL, [width, height, focal_length, width / 2, height / 2, radial_distortion])
        cameras[camera_id] = camera

        images[(timestamp, camera_id)] = image_file_name
        if trajectories is not None:
            trajectories[(timestamp, camera_id)] = pose
    return image_idx_to_image_name


def parse_points3d(kapture_path: str,
                   number_of_points: int,
                   nvm_content: List[str],
                   offset: int,
                   point_id_offset: int,
                   image_idx_to_image_name: List[str],
                   filter_list: Optional[Set[str]],
                   points3d: List[List[float]],
                   keypoints: kapture.Keypoints,
                   observations: kapture.Observations) -> None:
    """
    Parse the <List of points> section
    Fill points3d, keypoints, observations in place
    Write keypoints to disk.

    :param kapture_path: path to kapture root directory.
    :param number_of_points: number of points to process
    :param nvm_content: content of NVM file
    :param offset: number of characters to skip while reading every line
    :param point_id_offset:
    :param image_idx_to_image_name: list of images in their index order
    :param filter_list: optional list of images to process
    :param points3d: list of 3D points to extend
    :param keypoints: kapture keypoints list to extend
    :param observations: kapture observations to extend
    """
    # (image_name, nvm_feature_id ) -> keypoint_id
    known_keypoints = {}
    local_keypoints = {}
    for i in range(0, number_of_points):
        fields = nvm_content[i + offset].split()
        points3d.append([float(v) for v in fields[0:6]])
        # parse observations
        number_of_measurements = int(fields[6])
        for j in range(0, number_of_measurements):
            # parse measurement
            image_index = int(fields[7 + 4 * j + 0])
            feature_index = int(fields[7 + 4 * j + 1])
            x = float(fields[7 + 4 * j + 2])
            y = float(fields[7 + 4 * j + 3])

            # retrieve filename. if added, then proceed to add features / observations
            file_name = image_idx_to_image_name[image_index]
            if filter_list is not None and file_name not in filter_list:
                # file_name is not in the list, do not add it
                continue

            # init local_keypoints if needed
            if file_name not in local_keypoints:
                local_keypoints[file_name] = []
            # do not add the same keypoint twice
            if (file_name, feature_index) not in known_keypoints:
                # in the kapture format, keypoint id is different. Note that it starts from 0
                known_keypoints[(file_name, feature_index)] = len(local_keypoints[file_name])
                local_keypoints[file_name].append([x, y])
            keypoint_idx = known_keypoints[(file_name, feature_index)]
            point3d_idx = i + point_id_offset
            observations.add(point3d_idx, LOCAL_FEATURE_TYPE, file_name, keypoint_idx)

    # finally, convert local_keypoints to np.ndarray and add them to the global keypoints variable
    for image_filename, keypoints_array in local_keypoints.items():
        keypoints_np_array = np.array(keypoints_array, dtype=np.float32)
        keypoints_filepath = kapture.io.features.get_keypoints_fullpath(LOCAL_FEATURE_TYPE,
                                                                        kapture_path,
                                                                        image_filename)
        kapture.io.features.image_keypoints_to_file(keypoints_filepath, keypoints_np_array)
        keypoints.add(image_filename)


def import_nvm_command_line() -> None:
    """
    Do the NVM import to kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='import nvm file to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help='input path to nvm file')
    parser.add_argument('-im', '--images', default=None,
                        help='path to images directory.')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    parser.add_argument('--filter-list', default="",
                        help=('path to the filter list (optional), '
                              'this file contains a list a images, '
                              'one per line that will be kept in the output. '
                              'images not in the list will be skipped'))
    parser.add_argument('--ignore-trajectories', action='store_true', default=False,
                        help='Do not export extrinsics.')
    parser.add_argument('--add-reconstruction', action='store_true', default=False,
                        help='add the 3d points/keypoints/observations to the output')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_nvm(args.input, args.images, args.output, args.filter_list,
               args.ignore_trajectories, args.add_reconstruction, args.force, args.image_transfer)


if __name__ == '__main__':
    import_nvm_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

from copy import deepcopy
from typing import Set, Any
import numpy as np


class ImageFeaturesBase(Set[str]):
    """
    base class for per image features (keypoints, descriptors, global_features)
    """

    # features = set(path_to_image_file)
    # features.feature_filenames = set(path_to_feature_file)
    def __init__(self,
                 type_name: str,
                 dtype: Any,
                 dsize: int,
                 *args, **kwargs):
        """

        :param type_name: input name of feature type (eg. SIFT). Used for information only.
        :param dtype: input type of the feature (eg. numpy.float32).
        :param dsize: input length of the feature (eg. 64 for sift descriptor).
        :param args: anything that can be passed to the set() constructor.
        :param kwargs: anything that can be passed to the set() constructor.
        """
        if not isinstance(type_name, str):
            raise TypeError('Expect str as name')
        if not isinstance(dtype, (type, np.dtype)):
            raise TypeError('Expect type (or numpy.dtype) as dtype')
        if not isinstance(dsize, int):
            raise TypeError('Expect int as dsize')
        self._tname = type_name
        self._dtype = dtype
        self._dsize = dsize
        super().__init__(*args, **kwargs)

    @property
    def type_name(self) -> str:
        """
         :return: type name as string
        """
        return self._tname

    @property
    def dtype(self) -> type:
        """
        :return: feature type
        """
        return self._dtype

    @property
    def dsize(self) -> int:
        """
        :return: feature size
        """
        return self._dsize

    def __repr__(self):
        representation = f'{self.type_name} ({self.dtype.__name__} x {self.dsize}) = '
        if len(self) == 0:
            representation += '[]'
        else:
            representation += '[\n' + ',\n'.join(f'\t{i}' for i in self) + '\n]'
        return representation

    def __copy__(self):
        result = type(self)(self.type_name, self.dtype, self.dsize)
        result.update(self)
        return result

    def __deepcopy__(self, memo):
        result = type(self)(self.type_name, self.dtype, self.dsize)
        for v in self:
            result.add(deepcopy(v))
        return result


class Keypoints(ImageFeaturesBase):
    """
    Key points features
    """
    pass


class Descriptors(ImageFeaturesBase):
    """
    Image descriptor features
    """

    def __init__(self,
                 type_name: str,
                 dtype: Any,
                 dsize: int,
                 keypoints_type: str,
                 metric_type: str,
                 *args, **kwargs):
        """


        :param type_name: input name of feature type (eg. SIFT). Used for information only.
        :param dtype: input type of the feature (eg. numpy.float32).
        :param dsize: input length of the feature (eg. 64 for sift descriptor).
        :param keypoints_type: type of keypoints, name of the keypoints subfolder
        :param metric_type: name of the metric used to compare the descriptors, for information only.
        :param args: anything that can be passed to the set() constructor.
        :param kwargs: anything that can be passed to the set() constructor.
        """
        super().__init__(type_name, dtype, dsize, *args, **kwargs)
        if not isinstance(keypoints_type, str):
            raise TypeError('Expect str as keypoints_type')
        if not isinstance(metric_type, str):
            raise TypeError('Expect str as metric_type')
        self._keypoints_type = keypoints_type
        self._metric_type = metric_type

    @property
    def keypoints_type(self) -> str:
        """
        :return: keypoints_type
        """
        return self._keypoints_type

    @property
    def metric_type(self) -> str:
        """
        :return: metric_type
        """
        return self._metric_type

    def __repr__(self):
        representation = f'{self.type_name} {self.keypoints_type} {self.metric_type}' \
                         f' ({self.dtype.__name__} x {self.dsize}) = '
        if len(self) == 0:
            representation += '[]'
        else:
            representation += '[\n' + ',\n'.join(f'\t{i}' for i in self) + '\n]'
        return representation

    def __copy__(self):
        result = type(self)(self.type_name, self.dtype, self.dsize, self.keypoints_type, self.metric_type)
        result.update(self)
        return result

    def __deepcopy__(self, memo):
        result = type(self)(self.type_name, self.dtype, self.dsize, self.keypoints_type, self.metric_type)
        for v in self:
            result.add(deepcopy(v))
        return result


class GlobalFeatures(ImageFeaturesBase):
    """
    Global features
    """

    def __init__(self,
                 type_name: str,
                 dtype: Any,
                 dsize: int,
                 metric_type: str,
                 *args, **kwargs):
        """


        :param type_name: input name of feature type (eg. SIFT). Used for information only.
        :param dtype: input type of the feature (eg. numpy.float32).
        :param dsize: input length of the feature (eg. 64 for sift descriptor).
        :param metric_type: name of the metric used to compare the descriptors, for information only.
        :param args: anything that can be passed to the set() constructor.
        :param kwargs: anything that can be passed to the set() constructor.
        """
        super().__init__(type_name, dtype, dsize, *args, **kwargs)
        if not isinstance(metric_type, str):
            raise TypeError('Expect str as metric_type')
        self._metric_type = metric_type

    @property
    def metric_type(self) -> str:
        """
        :return: metric_type
        """
        return self._metric_type

    def __repr__(self):
        representation = f'{self.type_name} {self.metric_type} ({self.dtype.__name__} x {self.dsize}) = '
        if len(self) == 0:
            representation += '[]'
        else:
            representation += '[\n' + ',\n'.join(f'\t{i}' for i in self) + '\n]'
        return representation

    def __copy__(self):
        result = type(self)(self.type_name, self.dtype, self.dsize, self.metric_type)
        result.update(self)
        return result

    def __deepcopy__(self, memo):
        result = type(self)(self.type_name, self.dtype, self.dsize, self.metric_type)
        for v in self:
            result.add(deepcopy(v))
        return result



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a virtual gallery to the kapture format.
Virtual Gallery is a dataset produced by NaverLabs Europe.
It is available at https://europe.naverlabs.com/research/3d-vision/virtual-gallery-dataset/
It is subject to the Creative Commons Attribution Non Commercial No Derivatives 4.0 International License.

    Virtual gallery v10 look like this:
    v10 # input_root
    ├── training
        └── gallery_lightX_loopY
            └── extrinsic.txt # extrinsics for each image
            └── intrinsic.txt # intrinsics for each image, always the same
            └── frames
                 └── rgb
                     └── camera_Z
                            └── rgb_00229.jpg  # images
    ├── testing/
        └── gallery_lightX_occlusionW
            └── extrinsic.txt # extrinsics for each image
            └── intrinsic.txt # intrinsics for each image, always the same
            └── frames
                 └── rgb
                     └── camera_0
                            └── rgb_00229.jpg  # images
"""

import logging
import argparse
# kapture
import path_to_kapture  # noqa: F401
import kapture.utils.logging
from kapture.converter.virtual_gallery.import_virtual_gallery import import_virtual_gallery
from kapture.io.records import TransferAction

logger = logging.getLogger('virtual_gallery')


def import_virtual_gallery_command_line() -> None:
    """
    Do the virtual galley to kapture import using the command line parameters provided by the user.
    """
    parser = argparse.ArgumentParser(description='import virtual_gallery to kapture format')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)

    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True, help=('input path to virtual_gallery '
                                                              'v10 root directory'))
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('-c', '--configuration', default="all", choices=["training", "testing", "all"],
                        help='training, testing or all (both)')
    parser.add_argument('--light-range', nargs='+', default=list(range(1, 7)), type=int,
                        help='list of lights to include')
    parser.add_argument('--loop-range', nargs='+', default=list(range(1, 6)), type=int,
                        help='list of training loops to include')
    parser.add_argument('--camera-range', nargs='+', default=list(range(0, 6)), type=int,
                        help='list of training cameras to include')
    parser.add_argument('--occlusion-range', nargs='+', default=list(range(1, 5)), type=int,
                        help='list of testing occlusion levels to include')
    parser.add_argument('--as-rig', action='store_true', default=False,
                        help='in training trajectories, writes the position of the rig instead of individual cameras')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.skip,
                        help=f'How to import images [skip], '
                        f'choose among: {", ".join(a.name for a in TransferAction)}')
    ####################################################################################################################

    args = parser.parse_args()
    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_virtual_gallery(
        args.input,
        args.configuration,
        args.light_range,
        args.loop_range,
        args.camera_range,
        args.occlusion_range,
        args.as_rig,
        args.image_transfer,
        args.output,
        args.force
    )


if __name__ == '__main__':
    import_virtual_gallery_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Internal structure related operations.
"""

import os
import os.path as path
from shutil import rmtree
from typing import List, Optional
from .features import FEATURES_DATA_DIRNAMES
from .records import RECORD_DATA_DIRNAME
from .csv import CSV_FILENAMES
from kapture.utils.logging import getLogger
from kapture.utils.paths import path_secure


def delete_existing_kapture_files(
        dirpath: str,
        force_erase: bool,
        only: Optional[List[type]] = None,
        skip: Optional[List[type]] = None):
    """
    Deletes all existing files / directories at dirpath that corresponds to kapture data.
    do not use only and skip at the same time.

    :param dirpath:
    :param force_erase: do not ask user confirmation.
    :param only: can be used to select files / directories to be removed.
    :param skip: can be used to select files / directories to be kept.
    :return:
    """
    assert only is None or isinstance(only, list)
    assert skip is None or isinstance(skip, list)

    dirpath = path_secure(dirpath)
    csv_filepaths = [
        path.join(dirpath, filename)
        for dtype, filename in CSV_FILENAMES.items()
        if (not only and not skip) or (only and dtype in only) or (skip and dtype not in skip)]
    features_dirpaths = [
        path.join(dirpath, dirname)
        for dtype, dirname in FEATURES_DATA_DIRNAMES.items()
        if (not only and not skip) or (only and dtype in only) or (skip and dtype not in skip)]
    records_dirpaths = [RECORD_DATA_DIRNAME]
    # remove existing_files files (start with deepest/longest paths to avoid to delete files before dirs).
    existing_paths = list(reversed(sorted({pathval
                                           for pathval in csv_filepaths + features_dirpaths + records_dirpaths
                                           if path.isfile(pathval) or path.isdir(pathval)})))
    # if any
    if existing_paths:
        existing_paths_as_string = ', '.join(f'"{path.relpath(p, dirpath)}"' for p in existing_paths)
        # ask for permission
        to_delete = (force_erase or
                     (input(f'{existing_paths_as_string} already in "{dirpath}".'
                            ' Delete ? [y/N]').lower() == 'y'))

        # delete all or quit
        if to_delete:
            getLogger().info('deleting already'
                             f' existing {existing_paths_as_string}')
            for pathval in existing_paths:
                if path.islink(pathval) or path.isfile(pathval):
                    os.remove(pathval)
                else:
                    rmtree(pathval)
        else:
            raise ValueError(f'{existing_paths_as_string} already exist')
